<?php
/**
*	@file MySQLDB.class.php
*	@brief A Class for handling MySQL database operations
*/

/** 
 * A Class for handling MySQL database operations.
 * Needs to be inherited by class DB in order to password related stuff to work.
 */
class MySQLDB {
	/** MySQLDB $instance of MySQL database class. */
	private static $instance = null;
	/** MySQLDB MySQLDB::$db Database instance. */
	private $db;
	/** 
	 * bool MySQLDB::$debug Tells if we are debugging this code.
	 * If true will print error messages in case of errors and
	 *	while debugging it's OK to leak server setup details.
	 */
	private $debug = TRUE;

	private function __construct() {
		
		$config = Config::getInstance();
		
		$this->db = new mysqli($config->db_host, $config->db_user, $config->db_pass, $config->db_name, $config->db_port);

		if (mysqli_connect_errno()) {
			$this->fail('MySQL connect', mysqli_connect_error());
		}
    	// change character set to utf8
		if (!$this->db->set_charset("utf8")) {
    		$this->fail('MySQL set_charset', $this->db->error);
		}
	}
	
	/**
	 * Prints the error message if debugging is enabled.
    * @param string $pub Type of error.
    * @param string $pvt Error message.
    * @return void
   */
	private function fail($pub, $pvt = '') {

		$msg = $pub;
		if ($this->debug && $pvt !== '') {
			$msg .= ": $pvt";
		}
		/* The $pvt debugging messages may contain characters that would need to be
		 * quoted if we were producing HTML output,
		 * but we're using text/plain here.  Also, $debug is meant to be disabled on
		 * a "production install" to avoid leaking server setup details. */
		exit("An error occurred ($msg).\n");
	}
	
	/**
	 * Gets the only instance of the database
    *
    * @return MySQLDB Database object
   */
	public static function getInstance() {
		if (is_null(self::$instance)) {
			self::$instance = new MySQLDB();
		}
		return self::$instance;
	}
	
	/**
	 * Creates new user.
    * @param string $username Name of the new user.
    * @param string $password Password of the new user.
    * @return double User id or in case of error -1
   */
	public function createUser($username, $password, $email='') {

		// First check if username is already taken
		if (!$stmt = $this->db->prepare('SELECT user_id FROM users WHERE username=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('s', $username)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($count);
    	$stmt->fetch();
    	$stmt->close();

		// If user is found return error -1
		if ($count > 0) {
			return -1;
		}
		$config = Config::getInstance();
		// Else create new user
		if (version_compare($config->php_version, '5.3.7') >= 0) {
			$hash = password_hash($password, PASSWORD_BCRYPT);
		}
		else {
			$hasher = new PasswordHash($hash_cost_log2, $hash_portable);
			$hash = $hasher->HashPassword($password);
			if (strlen($hash) < 20) {
				fail('createUser', 'Failed to hash new password');
			}
			unset($hasher);
		}
		
		if (!$stmt = $this->db->prepare('INSERT INTO users (username, password, email) VALUES (?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}

		if (!$stmt->bind_param('sss', $username, $hash, $email)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Deletes user.
    * @param double $user_id ID of the user to be deleted.
    * @return int Affected rows. 0 if no user deleted.
   */
	public function deleteUser($user_id) {

		if (!$stmt = $this->db->prepare('DELETE FROM users WHERE user_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $user_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		return $result;
	}
	
	/**
	 * Creates new password for user.
	 * @param $username Name of the user.
	 * @param $email Email address of the user.
	 * @return string|false New password. False if given $username and $email don't match.
	 */
	public function resetPassword($username, $email) {
		
		// Check that username and email match
		if (!$stmt = $this->db->prepare('SELECT user_id FROM users WHERE username=? AND email=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ss', $username, $email)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->store_result();
		$num_of_rows = $stmt->num_rows;
		$stmt->bind_result($user_id);
    	$stmt->fetch();
    	$stmt->free_result();
		$stmt->close();
		
		// If matching user isn't found return false
		if ($num_of_rows != 1) {
			return false;
		}
		// Create random 8 character long password
		$new_password = $this->randomPassword(8);
		
		$config = Config::getInstance();
				
		$pass_hash = '';
		// Hash new password
		if (version_compare($config->php_version, '5.3.7') >= 0) {
			$pass_hash = password_hash($new_password, PASSWORD_BCRYPT);
		}
		else {
			$hasher = new PasswordHash($hash_cost_log2, $hash_portable);
			$pass_hash = $hasher->HashPassword($new_password);
			if (strlen($pass_hash) < 20) {
				return false;
			}
			unset($hasher);
		}
		// Store the new password hash to database
		if (!$stmt = $this->db->prepare('UPDATE users set password=? WHERE user_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sd', $pass_hash, $user_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$affected_rows = $stmt->affected_rows;
		$stmt->close();
		
		if ($affected_rows != 1) {
			return false;
		}
		return $new_password;
	}
	
	/**
	 * Creates random password.
    * @param int $password_length Number of characters in password.
    * @return string Created password.
   */
	private function randomPassword($password_length) {
		$alphabet = "abcdefghijklmnopqrstuwxyzABCDEFGHIJKLMNOPQRSTUWXYZ0123456789";
		$pass = array(); //remember to declare $pass as an array
		$alphaLength = strlen($alphabet) - 1; //put the length -1 in cache
		for ($i = 0; $i < $password_length; $i++) {
			$n = rand(0, $alphaLength);
			$pass[] = $alphabet[$n];
		}
		return implode($pass); //turn the array into a string
	}

	/**
	 * Retrieves user information (username) based on given user id.
    * @param double $user_id ID of the user.
    * @return array|false Containing 'user_id', 'username' and 'email'. False if user isn't found.
   */
	public function getUser($user_id) {
		
		if (!$stmt = $this->db->prepare('SELECT username, email FROM users WHERE user_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $user_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->store_result();
		$num_of_rows = $stmt->num_rows;
		$stmt->bind_result($username, $email);
    	$stmt->fetch();
    	$stmt->free_result();
		$stmt->close();
		
		// If singleuser isn't found return false
		if ($num_of_rows != 1) {
			return false;
		}
		$user = array();
		$user['user_id'] = $user_id;
		$user['username'] = $username;
		$user['email'] = $email;
		
		return $user;
	}
	
	/**
	 * Get all registered users that are tied to given game_id.
    * @param double $game_id ID of the game.
    * @return array Containing 'username' and 'user_id' subarrays.
   */
	public function getAllUsers($game_id) {
		// Inner join
		if (!$stmt = $this->db->prepare('SELECT users.username, users.user_id FROM users INNER JOIN player_profiles ON users.user_id=player_profiles.user_id WHERE game_id=? GROUP BY users.user_id')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($username, $user_id);
		
		$users = array();
		
		while ($result = $stmt->fetch()) {
			$user = array();
			$user['username'] = $username;
			$user['user_id'] = $user_id;
			$users[] = $user;
		}
		$stmt->close;
		
		return $users;
	}
	
	/**
	 * Logs user into game.
    * @param string $username Name of the user.
    * @param string $password Password of the user.
    * @return array|int User array if success else error code.
    * Error codes: -1: given user isn't found, -2: wrong password
   */
	public function login($username, $password) {
		
		$user = $this->authenticateUser($username, $password);

		if ($user <= 0) {
			// There was an error;
			return $user;
		}
		$user_array = $this->getUser($user);
		
		if (!$user_array) {
			// User isn't found;
			return -1;
		}
		$_SESSION['user'] = $user_array;
		$_SESSION['isLogged'] = true;
		return $user_array;
	}
	
	/**
	 * Changes user's password.
    * @param string $username Name of the user.
    * @param string $old_password Current password of the user.
    * @param string $new_password New password of the user.
    * @return bool Was operation successful?
   */
	public function changePassword($username, $old_password, $new_password) {
		
		// First authenticate user
		$user = $this->authenticateUser($username, $old_password);

		if ($user <= 0) {
			// There was an error;
			return false;
		}
		$config = Config::getInstance();
		if (version_compare($config->php_version, '5.3.7') >= 0) {
			$hash = password_hash($new_password, PASSWORD_BCRYPT);
		}
		else {
			$hasher = new PasswordHash($hash_cost_log2, $hash_portable);
			$hash = $hasher->HashPassword($new_password);
			if (strlen($hash) < 20) {
				fail('changePassword', 'Failed to hash new password');
			}
			unset($hasher);
		}

		if (!$stmt = $this->db->prepare('UPDATE users set password=? WHERE user_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sd', $hash, $user)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$affected_rows = $stmt->affected_rows;
		$stmt->close();
		
		if ($affected_rows != 1) {
			return false;
		}
		return true;
	}
	
	/**
	 * Changes user's email address.
    * @param string $username Name of the user.
    * @param string $password Current password of the user.
    * @param string $email New email address of the user.
    * @return bool Was operation successful?
   */
	public function changeEmail($username, $password, $email) {
		
		
		// First authenticate user
		$user = $this->authenticateUser($username, $password);
		
		if ($user <= 0) {
			// There was an error;
			return false;
		}
		

		if (!$stmt = $this->db->prepare('UPDATE users set email=? WHERE user_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sd', $email, $user)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$affected_rows = $stmt->affected_rows;
		$stmt->close();
		
		if ($affected_rows != 1) {
			return false;
		}
		return true;
	}
	
	/**
	 * Retrieves information of games user has played.
    * @param double $user_id ID of the user.
    * @return array Containing user's games.
   */
	public function getUsersGames($user_id) {
		
		if (!$stmt = $this->db->prepare('SELECT games.game_id, games.name, games.max_cancellations, player_profiles.profile_id, player_profiles.cancellations, player_profiles.created, player_profiles.is_panic, player_profiles.game_ended FROM games INNER JOIN player_profiles ON games.game_id=player_profiles.game_id WHERE user_id=? ORDER BY player_profiles.created')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $user_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($id, $name, $max_cancellations, $profile_id, $cancellations, $created, $is_panic, $game_ended);
		
		$games = array();
		
		while ($result = $stmt->fetch()) {
			$game = array();
			$game['game_id'] = $id;
			$game['name'] = $name;
			$game['max_cancellations'] = $max_cancellations;
			$game['profile_id'] = $profile_id;
			$game['cancellations'] = $cancellations;
			$game['profile_created'] = $created;
			$game['is_panic'] = $is_panic;
			$game['game_ended'] = $game_ended;
			$games[] = $game;
		}
		$stmt->close();
		
		// Get played cards count too
		foreach ($games as &$game) {
			$game['played_cards'] = count($this->getPlayedCards($game['profile_id']));
		}
		return $games;
	}
	
	/**
	 * Creates a new game.
    * @param string $name Name of the game. Must be unique.
    * @param string $admin_id Main administrator of the game.
    * @return int ID of the created game or in case of error -1.
   */
	public function createGame($name, $admin_id) {
		
		// First check if game name is already taken
		if (!$stmt = $this->db->prepare('SELECT game_id FROM games WHERE name=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('s', $name)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($count);
    	$stmt->fetch();
    	$stmt->close();

		// If game is found return error -1
		if ($count > 0) {
			return -1;
		}
		
		$unique = false;
		$slug = friendly_url($name);
		$counter = 1;
		// make slug unique
		while (!$unique) {
			
			$new_slug = $slug;
			if ($counter > 1) {
				$new_slug .= $counter; 
			}
			$game_id = $this->getGameIdBySlug($new_slug);

			// If slug is found increase counter
			if ($game_id > 0) {
				$counter++;
			}
			else {
				$unique = true;
				$slug = $new_slug;
			}
		}
		
		if (!$stmt = $this->db->prepare('INSERT INTO games (name, main_admin_id, slug) VALUES (?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sds', $name, $admin_id, $slug)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		if ($this->db->insert_id > 0) {
			
			$connection = $this->addAdmin($admin_id, $this->db->insert_id);
			if ($connection < 1) {
				// Remove created game from db
				$this->deleteGame($this->db->insert_id);
				return -1;
			}
		}
		return $this->db->insert_id;
	}
	
	/**
	 * Deletes game and all that is related to it.
    * @param double $game_id ID of the game to be deleted.
    * @return int Number of games deleted. 0 if none.
   */
	public function deleteGame($game_id) {
		
		// Delete game
		if (!$stmt = $this->db->prepare('DELETE FROM games WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		// Delete texts
		if (!$stmt = $this->db->prepare('DELETE FROM game_text_translations WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Delete admin relationships
		if (!$stmt = $this->db->prepare('DELETE FROM admin_relationships WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Delete sudden scenarios and scenario conditions
		$scenarios = $this->getSuddenScenarios($game_id);
		
		foreach ($scenarios as $scenario) {
			if (!$stmt = $this->db->prepare('DELETE FROM scenario_conditions WHERE scenario_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $scenario['scenario_id'])) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
		if (!$stmt = $this->db->prepare('DELETE FROM sudden_scenarios WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Delete ending criteria and ending conditions
		$criteria = $this->getEndingCriteria($game_id);
		
		foreach ($criteria as $crit) {
			if (!$stmt = $this->db->prepare('DELETE FROM ending_conditions WHERE criteria_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $crit['criteria_id'])) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
		if (!$stmt = $this->db->prepare('DELETE FROM ending_criteria WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
			
		// Get profiles
		$profiles = $this->getAllProfiles($game_id);
		// Delete played cards
		foreach ($profiles as $profile) {
			
			if (!$stmt = $this->db->prepare('DELETE FROM played_cards WHERE profile_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $profile['profile_id'])) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
		
		// Delete profiles
		if (!$stmt = $this->db->prepare('DELETE FROM player_profiles WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Delete cards
		$cards = $this->getAllCards($game_id, 'en-US', false, true);
		
		foreach ($cards as $card) {		
			$this->deleteCard($card['card_id']);
		}
		
		return $result;
	}
	
	/**
	 * Updates game settings.
    * @param array $game Array containing new game settings.
    * @return void
   */
	public function updateGame($game) {
		
		// First make sure that slug is unique
		$unique = false;
		$slug = friendly_url($game['slug']);
		if (empty($slug)) {
			$slug = friendly_url($game['name']);
		}
		$counter = 1;
		// make slug unique
		while (!$unique) {
			
			$new_slug = $slug;
			if ($counter > 1) {
				$new_slug .= $counter; 
			}
			$game_id = $this->getGameIdBySlug($new_slug);

			// If slug is found on another game increase counter
			if ($game_id != $game['game_id'] && $game_id > 0) {
				$counter++;
			}
			else {
				$unique = true;
				$slug = $new_slug;
			}
		}
		// Parse visible stats string
		$visible_stats = '';
		$stat_string = implode(' @', $game['visible_stats']);
		$stat_string = trim($stat_string);
		if (!empty($stat_string)) {
			$visible_stats = '@' . $stat_string;
		}
		// Update game settings to db
		if (!$stmt = $this->db->prepare(
				'UPDATE games set name=?, main_admin_id=?, starting_card_id=?, panic_card_id=?, image_folder=?, slug=?, latest_card_id=?,
				use_tags=?, use_stats=?, use_expert_advice=?, use_sudden_scenarios=?, use_ending_criteria=?, use_login=?, use_name=?, multiple_languages=?, show_cards_ending=?, max_cancellations=?,
				background_color=?, text_color=?, link_color=?, link_hover_color=?, button_color=?, button_hover_color=?, background_tile=?, logo=?, visible_stats=?
				WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sdddssdiiiiiiiiiisssssssssd', $game['name'], $game['main_admin_id'], $game['starting_card_id'], $game['panic_card_id'],
				$game['image_folder'], $slug, $game['latest_card_id'], $game['use_tags'], $game['use_stats'], $game['use_expert_advice'],
				$game['use_sudden_scenarios'], $game['use_ending_criteria'], $game['use_login'], $game['use_name'], $game['multiple_languages'], $game['show_cards_ending'], $game['max_cancellations'],
				$game['background_color'], $game['text_color'], $game['link_color'], $game['link_hover_color'], $game['button_color'], $game['button_hover_color'], $game['background_tile'], $game['logo'], $visible_stats,
				$game['game_id'])) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Remove old scenarios and conditions
		$this->deleteSuddenScenarios($game['game_id']);
		// Remove old criteria and conditions
		$this->deleteEndingCriteria($game['game_id']);
		
		// Add new scenarios
		if (isset($game['sudden_scenarios'])) {
			
			foreach ($game['sudden_scenarios'] as $scenario) {
				
				if (isset($scenario['odds']) && isset($scenario['result']) && isset($scenario['ordinal']) && isset($scenario['conditions'])) {
					
					$scenario_id = $this->createSuddenScenario($game['game_id'], $scenario['odds'], $scenario['result'], $scenario['ordinal']);

					if ($scenario_id > 0) {
						// Create new conditions
						foreach ($scenario['conditions'] as $condition) {
							$this->createCondition($scenario_id, $condition['condition_text'], $condition['condition_logic'], $condition['ordinal'], 'scenario_conditions');
						}
					}
				}
			} 
		}

		// Add new criteria
		if (isset($game['ending_criteria'])) {
			
			foreach ($game['ending_criteria'] as $criteria) {
				
				if (isset($criteria['result']) && isset($criteria['ordinal']) && isset($criteria['conditions'])) {
					
					$criteria_id = $this->createEndingCriteria($game['game_id'], $criteria['result'], $criteria['ordinal']);
					if ($criteria_id > 0) {
						// Create new conditions
						foreach ($criteria['conditions'] as $condition) {
							$this->createCondition($criteria_id, $condition['condition_text'], $condition['condition_logic'], $condition['ordinal'], 'ending_conditions');
						}
					}
				}
			}
		}
	}
	
	/**
	 * Updates extra options of the game.
	 * @param double $game_id Id of the game.
    * @param array $extras Array containing new extras.
    * @return void
   */
	public function updateGameExtras($game_id, $extras) {
		
		// Serialize php array
		$extras_str = serialize($extras);
		
		// Update extras to db
		if (!$stmt = $this->db->prepare('UPDATE games set extras=? WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sd', $extras_str, $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Retrieves information of all games.
    * @return array Containing information of all games.
   */
	public function getAllGames() {
		
		if (!($result = $this->db->query('SELECT * FROM games'))) {
			$this->fail('MySQL query', $this->db->error);
		}
		
		$games = array();
		
		if ($result->num_rows > 0) {

			while ($row = $result->fetch_object()) {
			
				$game = array();
				$game['game_id'] = $row->game_id;
				$game['name'] = $row->name;
				$game['created'] = $row->created;
				$game['main_admin_id'] = $row->main_admin_id;
				$game['starting_card_id'] = $row->starting_card_id;
				$game['panic_card_id'] = $row->panic_card_id;
				$game['image_folder'] = $row->image_folder;
				$game['slug'] = $row->slug;
				$game['latest_card_id'] = $row->latest_card_id;
				$game['use_tags'] = $row->use_tags;
				$game['use_stats'] = $row->use_stats;
				$game['use_expert_advice'] = $row->use_expert_advice;
				$game['use_sudden_scenarios'] = $row->use_sudden_scenarios;
				$game['use_ending_criteria'] = $row->use_ending_criteria;
				$game['use_login'] = $row->use_login;
				$game['use_name'] = $row->use_name;
				$game['multiple_languages'] = $row->multiple_languages;
				$game['show_cards_ending'] = $row->show_cards_ending;
				$game['max_cancellations'] = $row->max_cancellations;
				$game['background_color'] = $row->background_color;
				$game['text_color'] = $row->text_color;
				$game['link_color'] = $row->link_color;
				$game['link_hover_color'] = $row->link_hover_color;
				$game['button_color'] = $row->button_color;
				$game['button_hover_color'] = $row->button_hover_color;
				$game['background_tile'] = $row->background_tile;
				$game['logo'] = $row->logo;
				
				$extras = unserialize($row->extras);
				
				if (empty($extras)) {
					
					global $default_language;
					$extras = array();
					$extras['default_language'] = $default_language;
					$extras['card_template_id'] = 0;
				}
				$game['extras'] = $extras;
				
				$game['visible_stats'] = array();
				$stat_string = $row->visible_stats;
				$visible_stats = explode('@', $stat_string);
				if (!empty($visible_stats)) {
					foreach ($visible_stats as &$stat) {
						$stat = trim($stat);
					}
					$game['visible_stats'] = $visible_stats;
				}
				$games[] = $game;
			}
    		$result->close();
		}
		
		return $games;
	}
	
	/**
	 * Get information of single game.
    * @param double $game_id ID of the game to retrieve.
    * @return array|false An array containing game's information or false if game isn't found.
   */
	public function getGame($game_id) {

		if (!$stmt = $this->db->prepare('SELECT * FROM games WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($id, $name, $created, $main_admin_id, $starting_card_id, $panic_card_id, $image_folder,
								 $slug, $latest_card_id, $use_tags, $use_stats, $use_expert_advice, $use_sudden_scenarios,
								 $use_ending_criteria, $use_login, $use_name, $multiple_languages, $show_cards_ending, $max_cancellations,
								 $background_color, $text_color, $link_color, $link_hover_color, $button_color, $button_hover_color, $background_tile, $logo, $visible_stats, $extras_str);
		
		$num_of_rows = 0;
		$game = array();
		
		while ($result = $stmt->fetch()) {
			$game['game_id'] = $id;
			$game['name'] = $name;
			$game['created'] = $created;
			$game['main_admin_id'] = $main_admin_id;
			$game['starting_card_id'] = $starting_card_id;
			$game['panic_card_id'] = $panic_card_id;
			$game['image_folder'] = str_replace('\\', '/', $image_folder);
			$game['slug'] = $slug;
			$game['latest_card_id'] = $latest_card_id;
			$game['use_tags'] = $use_tags;
			$game['use_stats'] = $use_stats;
			$game['use_expert_advice'] = $use_expert_advice;
			$game['use_sudden_scenarios'] = $use_sudden_scenarios;
			$game['use_ending_criteria'] = $use_ending_criteria;
			$game['use_login'] = $use_login;
			$game['use_name'] = $use_name;
			$game['multiple_languages'] = $multiple_languages;
			$game['show_cards_ending'] = $show_cards_ending;
			$game['max_cancellations'] = $max_cancellations;
			$game['background_color'] = $background_color;
			$game['text_color'] = $text_color;
			$game['link_color'] = $link_color;
			$game['link_hover_color'] = $link_hover_color;
			$game['button_color'] = $button_color;
			$game['button_hover_color'] = $button_hover_color;
			$game['background_tile'] = $background_tile;
			$game['logo'] = $logo;
			
			$extras = unserialize($extras_str);
				
			if (empty($extras)) {
					
				global $default_language;
				$extras = array();
				$extras['default_language'] = $default_language;
				$extras['card_template_id'] = 0;
			}
			$game['extras'] = $extras;
			
			$game['visible_stats'] = array();
			$stat_string = $visible_stats;
			$visible_stats = explode('@', $stat_string);
			if (!empty($visible_stats)) {
				foreach ($visible_stats as $stat) {
					$stat = trim($stat);
					if (!empty($stat)) {
						$game['visible_stats'][] = $stat;
					}
				}
			}
			$num_of_rows++;
		}
		if ($num_of_rows != 1) {
			return false;
		}
		$stmt->close();
		
		$game['sudden_scenarios'] = $this->getSuddenScenarios($game_id);
		$game['ending_criteria'] = $this->getEndingCriteria($game_id);
		
		return $game;
	}
	
	/**
	 * Creates or updates game texts translations.
    * @param double $game_id ID of the game.
    * @param array $texts An array containing texts of the game in given language.
    * @param string $lang Language code of these texts.
    * @param int $is_default Defines whether or not these are the default texts true=1/false=0.
    * @param string $default_lang Language code of default texts.
    * @return void
   */
	public function addOrUpdateGameTexts($game_id, $texts, $lang, $is_default, $default_lang) {
		
		// First check if translation exists
		if (!$stmt = $this->db->prepare('SELECT translation_id FROM game_text_translations WHERE game_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $game_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($trans_id);
    	$stmt->fetch();
    	$trans_id = $trans_id;
		$stmt->close();
		
		// If not create one
		if ($trans_id < 1) {
			
			if (!$stmt = $this->db->prepare('INSERT INTO game_text_translations (game_id, welcome, front_up, front_down, registration_up, registration_down, login_up, login_down, new_up, new_down, open_up, open_down, continue_up, continue_down, loading, your_life, menu, selection, language_code, is_default) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('dssssssssssssssssssi', $game_id, $texts['welcome'], $texts['front_up'], $texts['front_down'], $texts['registration_up'], $texts['registration_down'],
									$texts['login_up'], $texts['login_down'], $texts['new_up'], $texts['new_down'],
									$texts['open_up'], $texts['open_down'], $texts['continue_up'], $texts['continue_down'], $texts['loading'], $texts['your_life'], $texts['menu'], $texts['selection'],
								 	$lang, $is_default)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
		// else update
		else {
			
			if (!$stmt = $this->db->prepare('UPDATE game_text_translations set welcome=?, front_up=?, front_down=?, registration_up=?, registration_down=?, login_up=?, login_down=?, new_up=?, new_down=?, open_up=?, open_down=?, continue_up=?, continue_down=?, loading=?, your_life=?, menu=?, selection=?, is_default=? WHERE game_id=? AND language_code=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('sssssssssssssssssids', $texts['welcome'], $texts['front_up'], $texts['front_down'], $texts['registration_up'], $texts['registration_down'],
										$texts['login_up'], $texts['login_down'], $texts['new_up'], $texts['new_down'],
										$texts['open_up'], $texts['open_down'], $texts['continue_up'], $texts['continue_down'], $texts['loading'], $texts['your_life'], $texts['menu'], $texts['selection'],
								 		$is_default, $game_id, $lang)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
		
		// Mark is_default false all but default
		if (!$stmt = $this->db->prepare('UPDATE game_text_translations set is_default=0 WHERE game_id=? AND language_code <> ?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $game_id, $default_lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark right translation to default
		if (!$stmt = $this->db->prepare('UPDATE game_text_translations set is_default=1 WHERE game_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $game_id, $default_lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Retrieves texts to application pages.
    * @param double $game_id ID of the game.
    * @return array Containing application texts.
   */
	public function getGameTexts($game_id) {
		
		if (!$stmt = $this->db->prepare('SELECT * FROM game_text_translations WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		
		$stmt->bind_result($translation_id, $game, $welcome, $front_up, $front_down, $registration_up, $registration_down, $login_up, $login_down,
								 $new_up, $new_down, $open_up, $open_down, $continue_up, $continue_down, $loading, $your_life, $menu, $selection,
								 $language_code, $is_default);
		
		$texts = array();
		
		while ($result = $stmt->fetch()) {
						
			$translation = array();
			$translation['translation_id'] = $translation_id;
			$translation['game_id'] = $game;
			$translation['welcome'] = html_entity_decode($welcome, ENT_COMPAT,'UTF-8');
			$translation['front_up'] = html_entity_decode($front_up, ENT_COMPAT,'UTF-8');
			$translation['front_down'] = html_entity_decode($front_down, ENT_COMPAT,'UTF-8');
			$translation['registration_up'] = html_entity_decode($registration_up, ENT_COMPAT,'UTF-8');
			$translation['registration_down'] = html_entity_decode($registration_down, ENT_COMPAT,'UTF-8');
			$translation['login_up'] = html_entity_decode($login_up, ENT_COMPAT,'UTF-8');
			$translation['login_down'] = html_entity_decode($login_down, ENT_COMPAT,'UTF-8');
			$translation['new_up'] = html_entity_decode($new_up, ENT_COMPAT,'UTF-8');
			$translation['new_down'] = html_entity_decode($new_down, ENT_COMPAT,'UTF-8');
			$translation['open_up'] = html_entity_decode($open_up, ENT_COMPAT,'UTF-8');
			$translation['open_down'] = html_entity_decode($open_down, ENT_COMPAT,'UTF-8');
			$translation['continue_up'] = html_entity_decode($continue_up, ENT_COMPAT,'UTF-8');
			$translation['continue_down'] = html_entity_decode($continue_down, ENT_COMPAT,'UTF-8');
			$translation['loading'] = html_entity_decode($loading, ENT_COMPAT,'UTF-8');
			$translation['your_life'] = html_entity_decode($your_life, ENT_COMPAT,'UTF-8');
			$translation['menu'] = html_entity_decode($menu, ENT_COMPAT,'UTF-8');
			$translation['selection'] = html_entity_decode($selection, ENT_COMPAT,'UTF-8');
			$translation['language_code'] = $language_code;
			$translation['is_default'] = $is_default;
			$texts[] = $translation;
		}
    	$stmt->close();
    	
		return $texts;
	}
	
	/**
	 * Retrieves corresponding game id based on given slug.
    * @param string $slug Slug of the game.
    * @return double ID of the game linked to given slug.
   */
	public function getGameIdBySlug($slug) {
		
		if (!$stmt = $this->db->prepare('SELECT game_id FROM games WHERE slug=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('s', $slug)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($id);
		$stmt->fetch();
    	$stmt->close();
    	
    	$game_id = $id;
    	if (empty($game_id)) {
    		return 0;
    	}
    	return $game_id;
	}
	
	/**
	 * Starts new simulation.
    * @param double $game_id ID of the game.
    * @param string $username User name of the player.
    * @return array|false Game information in array or false if function failed.
   */
	public function startNewGame($game_id, $username) {
		
		if ($game_id > 0) {
			// Get game array
			$game = $this->getGame($game_id);
		
			if (empty($game) || empty($username)) {
				return false;
			}
			// Draw scenario (1-5)
			$scenario = mt_rand(1, 5);
			$profile_id = $this->createNewProfile($game_id, $username, $scenario);
			if ($profile_id > 0) {
				$game['profile'] = $profile_id;
				return $game;
			}
		}
		return false;
	}
	
	/**
	 * Opens previously created game profile.
    * @param double $profile_id ID of the profile to be opened.
    * @return array|false Game information in array or false if function failed.
   */
	public function continueGame($profile_id) {
		
		if ($profile_id > 0) {
			$profile = $this->getProfile($profile_id);
			// Get game array
			$game = $this->getGame($profile['game_id']);
			
			if (empty($game)){
				return false;
			}
			$game['profile'] = $profile_id;
			return $game;
		}
		return false;
	}
	
	/**
	 * Adds new administrator for the game.
    * @param double $user_id ID of the admin to be added.
    * @param double $game_id ID of the game of which to add admin.
    * @return double|void ID of the admin relationship or void if relationship already exists.
   */
	public function addAdmin($user_id, $game_id) {
		
		// First check if connection already exists
		if (!$stmt = $this->db->prepare('SELECT * FROM admin_relationships WHERE user_id=? AND game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dd', $user_id, $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->store_result();
		$num_of_rows = $stmt->num_rows;
    	$stmt->fetch();
    	$stmt->free_result();
		$stmt->close();
		
		// If relationship is already found return
		if ($num_of_rows > 0) {
			return;
		}
		if (!$stmt = $this->db->prepare('INSERT INTO admin_relationships (user_id, game_id) VALUES (?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dd', $user_id, $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Removes an administrator of the game.
    * @param double $user_id ID of the admin to be removed.
    * @param double $game_id ID of the game of which to remove admin.
    * @return int Number of rows affected.
   */
	public function removeAdmin($user_id, $game_id) {

		if (!$stmt = $this->db->prepare('DELETE FROM admin_relationships WHERE user_id=? AND game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dd', $user_id, $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		return $result;
	}
	
	/**
	 * Retrieves all user ids of administrators of the game.
    * @param double $game_id ID of the game of which administrators to retrieve.
    * @return array An array of admin ids.
   */
	public function getAdminIDs($game_id) {
		
		if (!$stmt = $this->db->prepare('SELECT user_id FROM admin_relationships WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($user_id);
		
		$admins = array();
		
		while ($result = $stmt->fetch()) {
			$admin = array();
			$admins[] = $user_id;
		}
		$stmt->close();
		
		return $admins;
	}
	
	/**
	 * Retrieves all tags and stats related to the cards of the game.
    * @param double $game_id ID of the game of which tags to retrieve.
    * @return array An array containing 'tags' and 'stats'.
   */
	public function getAllTagsAndStats($game_id) {
		
		$current_lang = '';
		if (isset($_SESSION['lang'])) {
			$current_lang = $_SESSION['lang'];
		}
		else {
			global $language;
			$current_lang = $language;
		}
		$cards = $this->getAllCards($game_id, $current_lang, false, true);
		$tags_and_stats = array();
		$tags_and_stats['tags'] = array();
		$tags_and_stats['stats'] = array();
		
		foreach ($cards as $card) {
			
			foreach ($card['related_tags'] as $tag) {
				if (!in_array($tag, $tags_and_stats['tags'])) {
					$tags_and_stats['tags'][] = $tag;
				}
			}
			foreach ($card['related_stats'] as $stat) {
				if (!in_array($stat, $tags_and_stats['stats'])) {
					$tags_and_stats['stats'][] = $stat;
				}
			}
		}
		if (!empty($tags_and_stats['tags'])) {
			sort($tags_and_stats['tags'], SORT_STRING);
		}
		if (!empty($tags_and_stats['stats'])) {
			sort($tags_and_stats['stats'], SORT_STRING);
		}
		return $tags_and_stats;
	}
	
	/**
	 * Retrieves all tags related to the cards of the game.
    * @param double $game_id ID of the game of which tags to retrieve.
    * @return array An array of tags.
   */
	public function getAllTags($game_id) {
		
		$current_lang = '';
		if (isset($_SESSION['lang'])) {
			$current_lang = $_SESSION['lang'];
		}
		else {
			global $language;
			$current_lang = $language;
		}
		$cards = $this->getAllCards($game_id, $current_lang, false, true);
		$tags = array();
		
		foreach ($cards as $card) {
			
			foreach ($card['related_tags'] as $tag) {
				if (!in_array($tag, $tags)) {
					$tags[] = $tag;
				}
			}
		}
		if (!empty($tags)) {
			sort($tags, SORT_STRING);
		}
		return $tags;
	}
	
	/**
	 * Retrieves all stats related to the cards of the game.
    * @param double $game_id ID of the game of which stats to retrieve.
    * @return array An array of stats.
   */
	public function getAllStats($game_id) {
		
		$current_lang = '';
		if (isset($_SESSION['lang'])) {
			$current_lang = $_SESSION['lang'];
		}
		else {
			global $language;
			$current_lang = $language;
		}
		
		$cards = $this->getAllCards($game_id, $current_lang, false, true);
		$stats = array();
		
		foreach ($cards as $card) {
			
			foreach ($card['related_stats'] as $stat) {
				if (!in_array($stat, $stats)) {
					$stats[] = $stat;
				}
			}
		}
		
		if (!empty($stats)) {
			sort($stats, SORT_STRING);
		}
		return $stats;
	}
	
	/**
	 * Creates a new player profile.
    * @param double $game_id ID of the game which to create a new profile.
    * @param string $username The user name of a new profile.
    * @param int $scenario ID of the scenario.
    * @return int|false ID of the created profile or false in case of error.
   */
	public function createNewProfile($game_id, $username, $scenario) {
		
		if ($game_id > 0 && !empty($username)) {
			
			$user_id = 0;
			if (isset($_SESSION['user']['user_id']) && isset($_SESSION['isLogged']) && $_SESSION['isLogged']) {
				$user_id = $_SESSION['user']['user_id'];
			}
			
			if (!$stmt = $this->db->prepare('INSERT INTO player_profiles (game_id, user_id, player_name, scenario) VALUES (?, ?, ?, ?)')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('ddsi', $game_id, $user_id, $username, $scenario)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		
			return $this->db->insert_id;
		}
		return false;
	}
	
	/**
	 * Deletes a player profile.
    * @param double $profile_id ID of the player profile to be deleted.
    * @return int Affected database rows. Note: If there are no played cards returns 0.
   */
	public function deleteProfile($profile_id) {
		
		// Delete profile
		if (!$stmt = $this->db->prepare('DELETE FROM player_profiles WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $profile_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		if (!$result) {
			return $result;
		}
		// Delete played cards
		if (!$stmt = $this->db->prepare('DELETE FROM played_cards WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $profile_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		return $result;
	}
	
	/**
	 * Retrieves all player profiles related to game.
    * @param double $game_id ID of the game of which profiles to retrieve.
    * @return array An array of player profiles.
   */
	public function getAllProfiles($game_id) {
		
		if (!$stmt = $this->db->prepare('SELECT * FROM player_profiles WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($profile_id, $game, $user_id, $player_name, $created, $cancellations, $is_panic, $game_ended, $scenario);
		
		$profiles = array();
		
		while ($result = $stmt->fetch()) {
			$profile = array();
			$profile['profile_id'] = $profile_id;
			$profile['game_id'] = $game;
			$profile['user_id'] = $user_id;
			$profile['player_name'] = $player_name;
			$profile['created'] = $created;
			$profile['cancellations'] = $cancellations;
			$profile['is_panic'] = $is_panic;
			$profile['game_ended'] = $game_ended;
			$profile['scenario'] = $scenario;
			$profiles[] = $profile;
		}
		$stmt->close;
		
		// Get played cards count too
		foreach ($profiles as &$profile) {
			$profile['played_cards'] = count($this->getPlayedCards($profile['profile_id']));
		}
		
		return $profiles;
	}
	
	/**
	 * Retrieves a single player profile.
    * @param double $profile_id ID of the player profile to be retrieved.
    * @return array An array of information about player profile.
   */
	public function getProfile($profile_id) {
		
		if (!$stmt = $this->db->prepare('SELECT * FROM player_profiles WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $profile_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->store_result();
		$stmt->bind_result($id, $game_id, $user_id, $player_name, $created, $cancellations, $is_panic, $game_ended, $scenario);
		
		$profile = array();
		
		if ($result = $stmt->fetch()) {

			$profile['profile_id'] = $id;
			$profile['game_id'] = $game_id;
			$profile['user_id'] = $user_id;
			$profile['player_name'] = $player_name;
			$profile['created'] = $created;
			$profile['cancellations'] = $cancellations;
			$profile['is_panic'] = $is_panic;
			$profile['game_ended'] = $game_ended;
			$profile['scenario'] = $scenario;
			
			// Set scenario if not set
			if ($scenario < 1) {
				// Draw scenario (1-5)
				$profile['scenario'] = mt_rand(1, 5);
				$this->updatePlayerProfileScenario($id, $profile['scenario']);
			}
		}
		$stmt->free_result();
		$stmt->close;
		
		// Get played cards count too
		$profile['played_cards'] = count($this->getPlayedCards($profile_id));
		
		return $profile;
	}
	
	/**
	 * Updates number of rewindings done to a player profile.
    * @param double $profile_id ID of the player profile to be updated.
    * @param int $cancellations Number of rewindings done.
    * @return void
   */
	public function updatePlayerProfileCancellations($profile_id, $cancellations) {
	
		if (!$stmt = $this->db->prepare('UPDATE player_profiles SET cancellations=? WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', intval($cancellations), doubleval($profile_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Links user id to player profile.
    * @param double $profile_id ID of the player profile to be updated.
    * @param double $user_id ID of the user to be linked with profile.
    * @return void
   */
	public function updatePlayerProfileUser($profile_id, $user_id) {
	
		if (!$stmt = $this->db->prepare('UPDATE player_profiles SET user_id=? WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dd', doubleval($user_id), doubleval($profile_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Sets value of panic flag to a player profile.
    * @param double $profile_id ID of the player profile to be updated.
    * @param int $is_panic 0: No panic, 1: panic.
    * @return void
   */
	public function updatePlayerProfilePanic($profile_id, $is_panic) {
	
		if (!$stmt = $this->db->prepare('UPDATE player_profiles SET is_panic=? WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', intval($is_panic), doubleval($profile_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Marks game ended or not to a player profile.
    * @param double $profile_id ID of the player profile to be updated.
    * @param int $game_ended 0: Game hasn't ended, 1: game is ended.
    * @return void
   */
	public function updatePlayerProfileGameEnded($profile_id, $game_ended) {
	
		if (!$stmt = $this->db->prepare('UPDATE player_profiles SET game_ended=? WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', intval($game_ended), doubleval($profile_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Sets value of scenario to a player profile.
    * @param double $profile_id ID of the player profile to be updated.
    * @param int $scenario (1-5)
    * @return void
   */
	public function updatePlayerProfileScenario($profile_id, $scenario) {
	
		if (!$stmt = $this->db->prepare('UPDATE player_profiles SET scenario=? WHERE profile_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', intval($scenario), doubleval($profile_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Updates information about card user has played.
    * @param double $profile_id ID of the player profile of which card is to be updated.
    * @param double $card_id ID of the card.
    * @param double $choice_id ID of the choice user has made. 0 means none.
    * @param double $result_id ID of the result related to choice. 0 means none.
    * @param int $ordinal_nbr Ordinal number of the played card.
    * @return void
   */
	public function updatePlayedCard($profile_id, $card_id, $choice_id, $result_id, $ordinal_nbr) {
		
		if (!$stmt = $this->db->prepare('UPDATE played_cards SET choice_id=?, result_id=? WHERE profile_id=? AND card_id=? AND ordinal=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ddddi', $choice_id, $result_id, $profile_id, $card_id, $ordinal_nbr)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Retrieves information about cards user has played.
    * @param double $profile_id ID of the player profile.
    * @return array An array of played cards.
   */
	public function getPlayedCards($profile_id) {
		
		if (!$stmt = $this->db->prepare('SELECT cards.* FROM played_cards cards WHERE cards.profile_id = ? ORDER BY cards.ordinal ASC')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $profile_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($relationship_id, $profile, $card_id, $choice_id, $result_id, $ordinal);
		
		$cards = array();
		
		while ($result = $stmt->fetch()) {
			
			$card = array();
			$card['relationship_id'] = $relationship_id;
			$card['profile_id'] = $profile;
			$card['card_id'] = $card_id;
			$card['choice_id'] = $choice_id;
			$card['result_id'] = $result_id;
			$card['ordinal'] = $ordinal;
			$cards[] = $card;
		}
		$stmt->close();
		
		return $cards;
	}
	
	/**
	 * Removes card user has played.
    * @param double $profile_id ID of the player profile of which card is to be removed.
    * @param double $card_id ID of the card to be removed.
    * @param int $ordinal_nbr Ordinal number of the played card to be removed.
    * @return void
   */
	public function removePlayedCard($profile_id, $card_id, $ordinal_nbr) {
		
		if (!$stmt = $this->db->prepare('DELETE FROM played_cards WHERE profile_id=? AND card_id=? AND ordinal=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ddi', doubleval($profile_id), doubleval($card_id), intval($ordinal_nbr))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
	}
	
	/**
	 * Ties a new card to a player profile as being played.
    * @param double $profile_id ID of the player profile of which to add played card.
    * @param double $card_id ID of the card to be played.
    * @param int $ordinal_nbr Ordinal number of the played card.
    * @return void
   */
	public function playCard($profile_id, $card_id, $ordinal_nbr) {
		
		if (!$stmt = $this->db->prepare('INSERT INTO played_cards (profile_id, card_id, ordinal) VALUES (?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ddi', $profile_id, $card_id, $ordinal_nbr)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		$card_relationship = $this->db->insert_id;
		if ($card_relationship < 1) {
			return false;
		}
		
		return true;
	}
	
	/**
	 * Creates a new card with base information.
    * @param double $card_external_id ID of the card to be shown in editor (different than actual ID of the card).
    * @param double $game_id ID of the game in which card is created.
    * @param int $repetitive Flag to tell whether or not card can be played multiple times.
    * @param double $creator_id ID of the user who created the card.
    * @param string $tags Tags of the card.
    * @param string $stat_requirements Stat requirements of a card.
    * @param string $comment Comment creator has added to card.
    * @param string $attachment Name of the image linked to card.
    * @param string $layout Tells which layout card uses. Default value is 'normal' (this is undocumented feature and effect of this is left for developer to decide)
    * @return double ID of the created card.
   */
	public function createCard($card_external_id, $game_id, $repetitive, $creator_id, $tags, $stat_requirements='', $comment='', $attachment='', $layout='normal') {
		
		// Create card (created must be NOW()
		if (!$stmt = $this->db->prepare('INSERT INTO cards (created, card_external_id, game_id, repetitive, creator_id, tags, stat_requirements, comment, attachment, layout) VALUES (NOW(), ?, ?, ?, ?, ?, ?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ddidsssss', $card_external_id, $game_id, $repetitive, $creator_id, $tags, $stat_requirements, $comment, $attachment, $layout)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		$card_id = $this->db->insert_id;
		
		// Update latest_card_id to game
		if ($card_id > 0) {
			
			// First get current value
			if (!$stmt = $this->db->prepare('SELECT latest_card_id FROM games WHERE game_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $game_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($id);
    		$stmt->fetch();
    		$latest_card_id = $id;
			$stmt->close();
			// And then update it if necessary
			if ($latest_card_id == $card_external_id) {
				return $card_id;
			}
			if ($latest_card_id < $card_external_id) {
				$latest_card_id = $card_external_id;
			}
			
			if (!$stmt = $this->db->prepare('UPDATE games SET latest_card_id=? WHERE game_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('dd', $latest_card_id, $game_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
			
			$this->updateCardRelatedTagsAndStats($card_id);
		}
		
		// Return card_id
		return $card_id;
	}
	
	/**
	 * Deletes a card and all related information.
    * @param double $card_id ID of the card to be deleted.
    * @return void
   */
	public function deleteCard($card_id) {
		
		// Delete card
		if (!$stmt = $this->db->prepare('DELETE FROM cards WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		$stmt->close();
		
		// Delete translations
		if ($result > 0) {
			
			if (!$stmt = $this->db->prepare('DELETE FROM card_translations WHERE card_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $card_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$result = $this->db->affected_rows;
			$stmt->close();
		}
		// Delete choices and results
		if ($result > 0) {
			
			// Get choice_ids
			if (!$stmt = $this->db->prepare('SELECT choice_id FROM choices WHERE card_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $card_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($choice_id);
		
			$choices = array();
		
			while ($result = $stmt->fetch()) {
				$choices[] = $choice_id;
			}
			$stmt->close();
			// Delete choices
			if (!$stmt = $this->db->prepare('DELETE FROM choices WHERE card_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $card_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$result = $this->db->affected_rows;
			
			$stmt->close();
			
			if (count($choices) > 0) {
				
				foreach ($choices as $choice) {
					// Delete choice translations
					if (!$stmt = $this->db->prepare('DELETE FROM choice_translations WHERE choice_id=?')) {
						$this->fail('MySQL prepare', $this->db->error);
					}
					if (!$stmt->bind_param('d', $choice)) {
						$this->fail('MySQL bind_param', $this->db->error);
					}
					if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->close();
					
					// Get result_ids
					if (!$stmt = $this->db->prepare('SELECT result_id FROM results WHERE choice_id=?')) {
						$this->fail('MySQL prepare', $this->db->error);
					}
					if (!$stmt->bind_param('d', $choice)) {
						$this->fail('MySQL bind_param', $this->db->error);
					}
					if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->bind_result($result_id);
		
					$results = array();
		
					while ($res = $stmt->fetch()) {
						$results[] = $result_id;
					}
					$stmt->close();
					
					// Delete results
					if (!$stmt = $this->db->prepare('DELETE FROM results WHERE choice_id=?')) {
						$this->fail('MySQL prepare', $this->db->error);
					}
					if (!$stmt->bind_param('d', $choice)) {
						$this->fail('MySQL bind_param', $this->db->error);
					}
					if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->close();
					
					if (count($results) > 0) {
						
						// Delete result translations
						foreach ($results as $res) {
							if (!$stmt = $this->db->prepare('DELETE FROM result_translations WHERE result_id=?')) {
								$this->fail('MySQL prepare', $this->db->error);
							}
							if (!$stmt->bind_param('d', $res)) {
								$this->fail('MySQL bind_param', $this->db->error);
							}
							if (!$stmt->execute()) {
								$this->fail('MySQL execute', $this->db->error);
							}
							$stmt->close();
						}
					}
				}
			}
		}
		// Delete card from played_cards
		if (!$stmt = $this->db->prepare('DELETE FROM played_cards WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		
		$stmt->close();
		
		return $result;
	}
	
	/**
	 * Imports card into game. If given external id exist replaces that card.
    * @param double $card_external_id ID of the card to be shown in editor (different than actual ID of the card).
    * @param int $repetitive Flag to tell whether or not card can be played multiple times.
    * @param string $tags Tags of the card.
    * @param string $stat_requirements Stat requirements of a card.
    * @param string $comment Comment creator has added to card.
    * @param string $attachment A name of the image linked to card.
    * @param string $layout Tells which layout card uses. Default value is 'normal' (this is undocumented feature and effect of this is left for developer to decide)
    * @param string $card_name A name of the card in default language.
    * @param string $description A description of the card in default language.
    * @param string $expert_advice An expert advice of the card in default language.
    * @param array $choices Choices of the card in default language.
    * @param array $results Results of the card in default language.
    * @param double $game_id ID of the game in which to import card.
    * @param string $lang Language code of default language of the card.
    * @param double $creator_id ID of the user who created the card.
    * @return double ID of the created card.
   */
	public function importCard($card_external_id, $repetitive, $tags, $stat_requirements, $comment, $attachment, $layout='normal', $card_name, $description, $expert_advice, $choices, $results, $game_id, $lang, $creator_id) {
		
		// If card external id is given then first check if card already exists
		$card_exist = false;
		$card_id = 0;
		
		if (doubleval($card_external_id) > 0) {
			
			if (!$stmt = $this->db->prepare('SELECT card_id FROM cards WHERE card_external_id=? AND game_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('dd', doubleval($card_external_id), doubleval($game_id))) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($id);
    		$stmt->fetch();
    		$stmt->close();

			// If card is found mark that
			if ($id > 0) {
				$card_exist = true;
				$card_id = $id;
			}
		}
		// If card exist update that
		if ($card_exist) {
			
			$this->editCard(doubleval($card_id), doubleval($card_external_id), $repetitive, doubleval($editor_id), $tags, $stat_requirements, $comment, $attachment, $layout, doubleval($game_id));
			
			if (!$stmt = $this->db->prepare('SELECT translation_id FROM card_translations WHERE card_id=? AND language_code=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('ds', doubleval($card_id), $lang)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($translation_id);
    		$stmt->fetch();
    		$stmt->close();
			// If translation exists update that
			if ($translation_id > 0) {
				$is_default = 1;
				$card = $this->getCard($card_id);
				if ($card['default_lang'] != $lang) {
					$is_default = 0;
				}
				$this->updateCardTranslation(doubleval($card_id), $card_name, $description, $expert_advice, $lang, $is_default, $card['default_lang']);
				$this->updateCardChoicesAndResults(doubleval($card_id), $choices, $results, $lang, $is_default, $card['default_lang']);
				//$this->updateCardTranslation(doubleval($card_id), $card_name, $description, $expert_advice, $lang, true, $lang);
				//$this->updateCardChoicesAndResults(doubleval($card_id), $choices, $results, $lang, true, $lang);
				// $this->updateCardRelatedTagsAndStats($card_id); 
			}
			// Else create new translation 
			else {
				$this->addCardTranslation(doubleval($card_id), $card_name, $description, $expert_advice, $lang, false);
			
				$card = $this->getCard($card_id);
				$choice_counter = 0;
				foreach ($choices as $choice) {
					if (isset($card['choices'][$choice_counter])) {
						$trans_id = $this->addChoiceTranslation($card['choices'][$choice_counter]['choice_id'], $choice['text'], $lang, 0, doubleval($card_id));
						$card_results = $this->getChoiceResults($card['choices'][$choice_counter]['choice_id']);
						$result_counter = 0;
						foreach ($results[$choice_counter] as $result) {
							if (isset($card_results[$result_counter])) {
								$trans_id = $this->addResultTranslation($card_results[$result_counter]['result_id'], $result['text'], $lang, 0, doubleval($card_id));
							}
							$result_counter++;
						}
					}
					$choice_counter++;
				}
				//$this->addCardChoicesAndResults(doubleval($card_id), $choices, $results, $lang, false);
				//$this->updateCardChoicesAndResults(doubleval($card_id), $choices, $results, $lang, false, $card['default_lang']);
				//$this->updateCardRelatedTagsAndStats($card_id);
			}
		}
		// else create new card
		else {
			
			$new_card_id = 0;	
			$new_card_id = $this->createCard(doubleval($card_external_id), doubleval($game_id), $repetitive, doubleval($creator_id), $tags, $stat_requirements, $comment, $attachment, $layout);
			
			if ($new_card_id > 0) {
				
				$this->addCardTranslation($new_card_id, $card_name, $description, $expert_advice, $lang, true);
				$this->addCardChoicesAndResults($new_card_id, $choices, $results, $lang, true);
				// Update latest_card_id to game
				// First get current value
				if (!$stmt = $this->db->prepare('SELECT latest_card_id FROM games WHERE game_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($game_id))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($latest_id);
    			$stmt->fetch();
    			$latest_card_id = $latest_id;
				$stmt->close();
				// And then update it if necessary
				if ($latest_card_id != doubleval($card_external_id)) {
			
					if ($latest_card_id < doubleval($card_external_id)) {
				
						$latest_card_id = doubleval($card_external_id);

						if (!$stmt = $this->db->prepare('UPDATE games SET latest_card_id=? WHERE game_id=?')) {
							$this->fail('MySQL prepare', $this->db->error);
						}
						if (!$stmt->bind_param('dd', $latest_card_id, $game_id)) {
							$this->fail('MySQL bind_param', $this->db->error);
						}
						if (!$stmt->execute()) {
							$this->fail('MySQL execute', $this->db->error);
						}
						$stmt->close();
					}
				}
			}
		}
	}
	
	/**
	 * Updates card's base information.
	 * @param double $card_id The actual ID of the card to be updated.
    * @param double $card_external_id ID of the card to be shown in editor (different than actual ID of the card).
    * @param int $repetitive Flag to tell whether or not card can be played multiple times.
    * @param double $editor_id ID of the user who edited the card.
    * @param string $tags Tags of the card.
    * @param string $stat_requirements Stat requirements of a card.
    * @param string $comment Comment creator has added to card.
    * @param string $attachment A name of the image linked to card.
    * @param string $layout Tells which layout card uses. Default value is 'normal' (this is undocumented feature and effect of this is left for developer to decide)
    * @param double $game_id ID of the game in which card is related.
    * @return void
   */
	public function editCard($card_id, $card_external_id, $repetitive, $editor_id, $tags, $stat_requirements, $comment, $attachment, $layout='normal', $game_id) {
		
		if (!$stmt = $this->db->prepare('UPDATE cards SET card_external_id=?, repetitive=?, editor_id=?, tags=?, stat_requirements=?, comment=?, attachment=?, layout=? WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('didsssssd', $card_external_id, $repetitive, $editor_id, $tags, $stat_requirements, $comment, $attachment, $layout, $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Update latest_card_id to game
		// First get current value
		if (!$stmt = $this->db->prepare('SELECT latest_card_id FROM games WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($id);
    	$stmt->fetch();
    	$latest_card_id = $id;
		$stmt->close();
		// And then update it if necessary
		if ($latest_card_id != $card_external_id) {
			
			if ($latest_card_id < $card_external_id) {
				
				$latest_card_id = $card_external_id;

				if (!$stmt = $this->db->prepare('UPDATE games SET latest_card_id=? WHERE game_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('dd', $latest_card_id, $game_id)) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->close();
			}
		}
		$this->updateCardRelatedTagsAndStats($card_id);
	}
	
	/**
	 * Sets or unsets card to be drawn only when it's ID is defined in criteria.
	 * @param double $card_id The actual ID of the card to be updated.
	 * @param int $hidden Flag 0: false, 1: true
	 * @return void
	 */
	function toggleCardVisibility($card_id, $hidden) {
		
		if (!$stmt = $this->db->prepare('UPDATE cards SET hidden=? WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', $hidden, $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Adds choices and results to card.
	 * @param double $card_id The actual ID of the card to be updated.
    * @param array $choices Choices to be added.
    * @param array $results Results to be added.
    * @param string $lang Language code of the choices and results.
    * @param int $is_default Flag that determines whether or not choices are in default language. 0: false, 1: true
    * @return void|false Returns false in case of errors.
   */
	public function addCardChoicesAndResults($card_id, $choices, $results, $lang, $is_default) {
		
		if ($card_id <= 0) {
			return false;
		}
		$choice_counter = 0;
		foreach ($choices as $choice) {
			
			$choice_id = $this->createChoice($card_id, $choice_counter+1);
			if ($choice_id > 0) {
				
				$choice_trans_id = $this->addChoiceTranslation($choice_id, $choice['text'], $lang, $is_default);
				
				$result_counter = 0;
				foreach ($results[$choice_counter] as $result) {
					
					$result_id = $this->createResult($choice_id, $result['weight'], $result['stat_changes'], $result['next_criteria'], $result_counter+1, $choice_counter+1);
					if ($result_id > 0) {
			
						$result_trans_id = $this->addResultTranslation($result_id, $result['text'], $lang, $is_default);
					}
					$result_counter++;
				}
				$choice_counter++;
			}
		}
		$this->updateCardRelatedTagsAndStats($card_id);
	}
	
	/**
	 * Updates choices and results of the card.
	 * @param double $card_id The actual ID of the card to be updated.
    * @param array $choices Choices to be updated.
    * @param array $results Results to be updated.
    * @param string $lang Language code of the choices and results.
    * @param int $is_default Flag that determines whether or not choices are in default language. 0: false, 1: true
    * @param string $default_lang The default language of the choices and results.
    * @return void|false Returns false in case of errors.
   */
	public function updateCardChoicesAndResults($card_id, $choices, $results, $lang, $is_default, $default_lang) {
		
		if ($card_id <= 0) {
			return false;
		}
		
		// Get old choices first
		if (!$stmt = $this->db->prepare('SELECT choice_id FROM choices WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($choice_id);
		
		$old_choices = array();
		
		while ($res = $stmt->fetch()) {
			$old_choices[] = $choice_id;
		}
		$stmt->close();
		
		// Get all card translations
		$translations = $this->getCardTranslations($card_id);
		
		$new_choices_count = count($choices);
		$limit = 0;
		// Remove choices and results that aren't needed anymore
		if ($new_choices_count < count($old_choices)) {
			$limit = $new_choices_count;
			for ($i = $new_choices_count; $i < count($old_choices); $i++) {
				
				// Delete choices
				if (!$stmt = $this->db->prepare('DELETE FROM choices WHERE card_id=? AND choice_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('dd', $card_id, $old_choices[$i])) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$res = $this->db->affected_rows;
				$stmt->close();
			
				// Delete choice translations
				if (!$stmt = $this->db->prepare('DELETE FROM choice_translations WHERE choice_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', $old_choices[$i])) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->close();
				
				// Get result_ids
				if (!$stmt = $this->db->prepare('SELECT result_id FROM results WHERE choice_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', $old_choices[$i])) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($result_id);
		
				$old_results = array();
		
				while ($res = $stmt->fetch()) {
					$old_results[] = $result_id;
				}
				$stmt->close();
					
				// Delete results
				if (!$stmt = $this->db->prepare('DELETE FROM results WHERE choice_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', $old_choices[$i])) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->close();
					
				if (count($old_results) > 0) {
						
					// Delete result translations
					foreach ($old_results as $res) {
						if (!$stmt = $this->db->prepare('DELETE FROM result_translations WHERE result_id=?')) {
							$this->fail('MySQL prepare', $this->db->error);
						}
						if (!$stmt->bind_param('d', $res)) {
							$this->fail('MySQL bind_param', $this->db->error);
						}
						if (!$stmt->execute()) {
							$this->fail('MySQL execute', $this->db->error);
						}
						$stmt->close();
					}
				}
			}
		}
		// Create new choices and results
		else if ($new_choices_count > count($old_choices)) {
					
			$limit = count($old_choices);
			for ($i = count($old_choices); $i < $new_choices_count; $i++) {

				$choice_id = $this->createChoice($card_id, $i+1);
				if ($choice_id > 0) {
					// Create translations for all languages
					foreach ($translations as $trans) {	
						$choice_trans_id = $this->addChoiceTranslation($choice_id, $choices[$i]['text'], $trans['language_code'], $trans['is_default']);
					}
					$result_counter = 0;
					foreach ($results[$i] as $result) {
					
						$result_id = $this->createResult($choice_id, $result['weight'], $result['stat_changes'], $result['next_criteria'], $result_counter+1, $i+1);
						if ($result_id > 0) {
							// Create translations for all languages
							foreach ($translations as $trans) {
								$result_trans_id = $this->addResultTranslation($result_id, $result['text'], $trans['language_code'], $trans['is_default']);
							}
							$result_counter++;
						}
					}
				}
			}
		}
		else {
			$limit = $new_choices_count;
		}
		
		// update old choices
		for ($i = 0; $i < $limit; $i++) {
			
			$choice_id = $old_choices[$i];
			$this->updateChoiceOrdinal($choice_id, $i+1);
			$this->updateChoiceTranslation($choice_id, $choices[$i]['text'], $lang, $is_default, $default_lang);
			
			// Get result_ids
			if (!$stmt = $this->db->prepare('SELECT result_id FROM results WHERE choice_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $choice_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($result_id);
		
			$old_results = array();
		
			while ($res = $stmt->fetch()) {
				$old_results[] = $result_id;
			}
			$stmt->close();
			
			$result_limit = 0;
			// Delete unused results
			if (count($results[$i]) < count($old_results)) {
				$result_limit = count($results[$i]);
				for ($j = count($results[$i]); $j < count($old_results); $j++) {
			
					if (!$stmt = $this->db->prepare('DELETE FROM results WHERE choice_id=? AND result_id=?')) {
						$this->fail('MySQL prepare', $this->db->error);
					}
					if (!$stmt->bind_param('dd', $choice_id, $old_results[$j])) {
						$this->fail('MySQL bind_param', $this->db->error);
					}
					if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->close();
					
					// Delete result translations
					if (!$stmt = $this->db->prepare('DELETE FROM result_translations WHERE result_id=?')) {
						$this->fail('MySQL prepare', $this->db->error);
					}
					if (!$stmt->bind_param('d', $old_results[$j])) {
						$this->fail('MySQL bind_param', $this->db->error);
					}
					if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->close();
				}
			}
			// Add new results
			elseif (count($results[$i]) > count($old_results)) {
				$result_limit = count($old_results);
				for ($j = count($old_results); $j < count($results[$i]); $j++) {

					$result_id = $this->createResult($choice_id, $results[$i][$j]['weight'], $results[$i][$j]['stat_changes'], $results[$i][$j]['next_criteria'], $j+1, $i+1);
					if ($result_id > 0) {
						// Create translations for all languages
						foreach ($translations as $trans) {	
							$result_trans_id = $this->addResultTranslation($result_id, $results[$i][$j]['text'], $trans['language_code'], $trans['is_default']);
						}
					}
				}
			}
			else {
				$result_limit = count($results[$i]);
			}
			
			// Update old results
			for ($j = 0; $j < $result_limit; $j++) {
				$this->updateResult($old_results[$j], $results[$i][$j]['weight'], $results[$i][$j]['stat_changes'], $results[$i][$j]['next_criteria'], $j+1, $i+1);
				
				$this->updateResultTranslation($old_results[$j], $results[$i][$j]['text'], $lang, $is_default, $default_lang);
			}
		}
		$this->updateCardRelatedTagsAndStats($card_id);
	}
	
	/**
	 * Retrieves information about single result.
	 * @param double $result_id ID of the result to retrieve.
    * @param string $lang Language code of the result to retrieve.
    * @return array An array containing the information about retrieved result.
   */
	public function getResult($result_id, $lang='en-US') {
		
		if (!$stmt = $this->db->prepare('SELECT r.* FROM results r WHERE r.result_id = ?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($result_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($res_id, $choice_id, $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal);
		
		$single_result = array();
		
		if ($result = $stmt->fetch()) {
			
			$single_result['result_id'] = $res_id;
			$single_result['choice_id'] = $choice_id;
			$single_result['weight'] = $weight;
			$single_result['stat_changes'] = $stat_changes;
			$single_result['next_criteria'] = $next_card_criteria;
			$single_result['ordinal'] = $ordinal;
			$single_result['choice_ordinal'] = $choice_ordinal;
		}
		$stmt->free_result();
		$stmt->close();
		
		if (!$stmt = $this->db->prepare('SELECT rt.result_text, rt.language_code, rt.is_default FROM result_translations rt WHERE rt.result_id = ?')) {
				$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($single_result['result_id']))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($result_text, $language_code, $is_default);
			
		$single_result['result_text'] = '';
		$single_result['lang'] = $lang;
		
		while ($result = $stmt->fetch()) {
				
			if ($language_code == $lang) {
				
				$single_result['result_text'] = $result_text;
			}
			if ($is_default) {
				$single_result['default_text'] = $result_text;
				$single_result['default_lang'] = $language_code;
			}
		}
		$stmt->free_result();
		$stmt->close();
		
		return $single_result;
	}
	
	/**
	 * Retrieves results linked to a single choice.
	 * @param double $choice_id ID of the choice of which results to retrieve.
    * @param string $lang Language code of the results to retrieve.
    * @param bool $include_translations Flag to tell if choice translations are wanted (default true).
    * @return array An array containing the results.
   */
	public function getChoiceResults($choice_id, $lang='en-US', $include_translations=true) {

		if (!$stmt = $this->db->prepare('SELECT r.* FROM results r WHERE r.choice_id = ? ORDER BY r.ordinal ASC')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($choice_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($result_id, $choi_id, $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal);
		
		$results = array();
		
		while ($result = $stmt->fetch()) {
			
			$single_result = array();
			$single_result['result_id'] = $result_id;
			$single_result['choice_id'] = $choi_id;
			$single_result['weight'] = $weight;
			$single_result['stat_changes'] = $stat_changes;
			$single_result['next_criteria'] = $next_card_criteria;
			$single_result['ordinal'] = $ordinal;
			$single_result['choice_ordinal'] = $choice_ordinal;
			$results[] = $single_result;
		}
		$stmt->free_result();
		$stmt->close();
		
		if ($include_translations) {
			foreach ($results as &$res) {
		
				if (!$stmt = $this->db->prepare('SELECT rt.result_text, rt.language_code, rt.is_default FROM result_translations rt WHERE rt.result_id = ?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($res['result_id']))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($result_text, $language_code, $is_default);
				
				$res['result_text'] = '';
				$res['lang'] = $lang;
			
				while ($result = $stmt->fetch()) {
					
					if ($language_code == $lang) {
						$res['result_text'] = $result_text;
					}
					if ($is_default) {
						$res['default_text'] = $result_text;
						$res['default_lang'] = $language_code;
					}
				}
				$stmt->free_result();
				$stmt->close();
			}
		}
		return $results;
	}
	
	/**
	 * Adds new translation of a result.
	 * @param double $result_id ID of the result in which to add new translation.
	 * @param string $result_text Text of the result in translation language.
	 * @param string $lang Language code of the new translation.
    * @param int $is_default Flag that determines whether or not new translation is in default language. 0: false, 1: true
    * @param double $card_id ID of the card which result is related.
    * @return double ID of the new translation or the old one if it already existed.
   */
	public function addResultTranslation($result_id, $result_text, $lang, $is_default, $card_id=0) {
		
		// First check if result already has translation in current language
		if (!$stmt = $this->db->prepare('SELECT translation_id FROM result_translations WHERE result_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $result_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($trans_id);
    	$stmt->fetch();
    	$stmt->close();

		// If translation is found return translation_id
		if ($trans_id > 0) {
			return $trans_id;
		}
		
		if (!$stmt = $this->db->prepare('INSERT INTO result_translations (result_id, result_text, language_code, is_default) VALUES (?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dssi', $result_id, $result_text, $lang, $is_default)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Adds new translation of a choice.
	 * @param double $choice_id ID of the choice in which to add new translation.
	 * @param string $choice_text Text of the choice in translation language.
	 * @param string $lang Language code of the new translation.
    * @param int $is_default Flag that determines whether or not new translation is in default language. 0: false, 1: true
    * @param double $card_id ID of the card which choice is related.
    * @return double ID of the new translation or the old one if it already existed.
   */
	public function addChoiceTranslation($choice_id, $choice_text, $lang, $is_default, $card_id=0) {
		
		// First check if choice already has translation in current language
		if (!$stmt = $this->db->prepare('SELECT translation_id FROM choice_translations WHERE choice_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $choice_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($trans_id);
    	$stmt->fetch();
    	$stmt->close();

		// If translation is found return translation_id
		if ($trans_id > 0) {
			return $trans_id;
		}
		
		if (!$stmt = $this->db->prepare('INSERT INTO choice_translations (choice_id, choice_text, language_code, is_default) VALUES (?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dssi', $choice_id, $choice_text, $lang, $is_default)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Adds new translation of a card.
	 * @param double $card_id ID of the card in which to add new translation.
	 * @param string $card_name Name of the card in translation language.
	 * @param string $description Description of the card in translation language.
	 * @param string $expert_advice Expert advice of the card in translation language.
	 * @param string $lang Language code of the new translation.
    * @param int $is_default Flag that determines whether or not new translation is in default language. 0: false, 1: true
    * @return double ID of the new translation or the old one if it already existed.
   */
	public function addCardTranslation($card_id, $card_name, $description, $expert_advice, $lang, $is_default) {
		
		// First check if card already has translation in current language
		if (!$stmt = $this->db->prepare('SELECT translation_id FROM card_translations WHERE card_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $card_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($trans_id);
    	$stmt->fetch();
    	$stmt->close();

		// If translation is found return translation_id
		if ($trans_id > 0) {
			return $trans_id;
		}
		
		if (!$stmt = $this->db->prepare('INSERT INTO card_translations (card_id, card_name, description, expert_advice, language_code, is_default) VALUES (?, ?, ?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dssssi', $card_id, $card_name, $description, $expert_advice, $lang, $is_default)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Updates translation of a card.
	 * @param double $card_id ID of the card in which translation to update.
	 * @param string $card_name Name of the card in translation language.
	 * @param string $description Description of the card in translation language.
	 * @param string $expert_advice Expert advice of the card in translation language.
	 * @param string $lang Language code of the new translation.
    * @param int $is_default Flag that determines whether or not new translation is in default language. 0: false, 1: true
    * @param string $default_lang Language code of the default language.
    * @return void
   */
	public function updateCardTranslation($card_id, $card_name, $description, $expert_advice, $lang, $is_default, $default_lang) {
		
		if (!$stmt = $this->db->prepare('UPDATE card_translations set card_name=?, description=?, expert_advice=?, is_default=? WHERE card_id=? and language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sssids', $card_name, $description, $expert_advice, $is_default, $card_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark is_default false to all translations
		if (!$stmt = $this->db->prepare('UPDATE card_translations set is_default=0 WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark right translation to default
		if (!$stmt = $this->db->prepare('UPDATE card_translations set is_default=1 WHERE card_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $card_id, $default_lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Retrieves all translations of a card.
	 * @param double $card_id ID of the card in which translations to retrieve.
    * @return array An array of translations.
   */
	public function getCardTranslations($card_id) {
		
		if (!$stmt = $this->db->prepare('SELECT ct.* FROM card_translations ct WHERE ct.card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($translation_id, $card, $card_name, $description, $expert_advice,
								 $language_code, $is_default);
		
		$translations = array();
		$counter = 0;
		while ($result = $stmt->fetch()) {
			$translation = array();
			$translation['translation_id'] = $translation_id;
			$translation['card_id'] = $card;
			$translation['card_name'] = $card_name;
			$translation['description'] = $description;
			$translation['expert_advice'] = $expert_advice;
			$translation['language_code'] = $language_code;
			$translation['is_default'] = $is_default;
			$translations[] = $translation;
			$counter++;
		}
		$stmt->close();

		return $translations;
	}
	
	/**
	 * Retrieves the number of all cards related to a game.
	 * @param double $game_id ID of the game of which cards to retrieve.
    * @return int Number of cards.
   */
	public function getNbrOfCardsInGame($game_id) {
		
		if (!$stmt = $this->db->prepare('SELECT card_id FROM cards WHERE game_id = ?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		
		$stmt->store_result();
		$num_of_rows = $stmt->num_rows;
    	$stmt->fetch();
    	$stmt->free_result();
		$stmt->close();
		
		return $num_of_rows;
	}
	
	/**
	 * Retrieves all cards related to a game.
	 * @param double $game_id ID of the game of which cards to retrieve.
	 * @param string $lang Language code of the language cards are wanted.
	 * @param bool $include_choices Flag to tell if card choices are fetched too (default true).
	 * @param bool $less Flag to tell if card translations, creator_name and editor_name aren't wanted (default false).
	 * @param int $limit Tells how many cards to return at max (default -1 == no limit).
	 * @param int $offset_id Ids of returned cards must be bigger than this (default 0 == no offset).
    * @return array An array of cards.
   */
	public function getAllCards($game_id, $lang, $include_choices=true, $less=false, $limit=-1, $offset_id=0) {
		
		ini_set('max_execution_time', 300); // 5 minutes
		set_time_limit(300);
		
//		$mtime = microtime();
//   	$mtime = explode(" ",$mtime);
//   	$mtime = $mtime[1] + $mtime[0];
//   	$starttime = $mtime; 
		
		$query_str = 'SELECT c.* FROM cards c WHERE c.game_id = ?';
		if ($offset_id > 0) {
			$query_str .= ' AND c.card_id > ' . intval($offset_id);
		}
		$query_str .= ' ORDER BY c.card_id ASC';
		if ($limit > -1) {
			$query_str .= ' LIMIT ' . intval($limit);
		}
		if (!$stmt = $this->db->prepare($query_str)) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $game_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		
		$stmt->bind_result($card_id, $card_external_id, $game, $enabled, $repetitive, $created, $modified,
								 $creator_id, $editor_id, $tags, $stat_requirements, $comment, $attachment, $layout,
								 $related_tags, $related_stats, $hidden);
		
		$cards = array();
		
		while ($result = $stmt->fetch()) {
			
			$card = array();
			$card['card_id'] = $card_id;
			$card['card_external_id'] = $card_external_id;
			if ($card['card_external_id'] < 1) {
				$card['card_external_id'] = $card['card_id'];
			}
			$card['game_id'] = $game;
			$card['enabled'] = $enabled;
			$card['repetitive'] = $repetitive;
			$card['created'] = $created;
			$card['modified'] = $modified;
			$card['creator_id'] = $creator_id;
			$card['editor_id'] = $editor_id;
			
			$tags_array = explode(' ', $tags);
			$clean_tags = array();
				
			foreach ($tags_array as $tag) {
				$trimmed = trim($tag);
				if (!empty($trimmed)) {
					$clean_tags[] = $trimmed;
				}
			}
			$card['tags'] = $clean_tags;
			
			$stats_array = explode(' ', $stat_requirements);
			$clean_stats = array();
				
			foreach ($stats_array as $stat) {
				$trimmed = trim($stat);
				if (!empty($trimmed)) {
					$clean_stats[] = $trimmed;
				}
			}
			$card['stat_requirements'] = $clean_stats;
			$card['comment'] = $comment;
			$card['attachment'] = $attachment;
			$card['layout'] = $layout;
			
			$card['related_tags'] = array();
			$card['related_stats'] = array();
			
			$tags_array = explode(' ', $related_tags);
			$clean_tags = array();
				
			foreach ($tags_array as $tag) {
				$trimmed = trim($tag);
				if (!empty($trimmed)) {
					$clean_tags[] = $trimmed;
				}
			}
			$card['related_tags'] = $clean_tags;
			
			$stats_array = explode(' ', $related_stats);
			$clean_stats = array();
				
			foreach ($stats_array as $stat) {
				$trimmed = trim($stat);
				if (!empty($trimmed)) {
					$clean_stats[] = $trimmed;
				}
			}
			$card['related_stats'] = $clean_stats;
			
			$card['hidden'] = $hidden;
			
			$cards[] = $card;
		}
		$stmt->free_result();
		$stmt->close(); 
   	
		if (!$less) {
			foreach ($cards as &$card) {
				
				if (!$stmt = $this->db->prepare('SELECT ct.card_name, ct.description, ct.expert_advice, ct.language_code, ct.is_default FROM card_translations ct WHERE ct.card_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', $card['card_id'])) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($card_name, $description, $expert_advice, $language_code, $is_default);
				
				$card['default_name'] = '';
				$card['default_description'] = '';
				$card['default_expert_advice'] = '';
				$card['default_lang'] = '';
				$card['card_name'] = '';
				$card['description'] = '';
				$card['expert_advice'] = '';
			
				while ($result = $stmt->fetch()) {
					
					if ($language_code == $lang) {
					
						$card['card_name'] = $card_name;
						$card['description'] = $description;
						$card['expert_advice'] = $expert_advice;
						$card['lang'] = $lang;
					}
					if ($is_default) {
						$card['default_name'] = $card_name;
						$card['default_description'] = $description;
						$card['default_expert_advice'] = $expert_advice;
						$card['default_lang'] = $language_code;
					}
				}
				if (!isset($card['lang'])) {
					$card['lang'] = $card['default_lang'];
				}
				$stmt->free_result();
				$stmt->close();
				
				$num_of_rows = 0;
	
				if (!$stmt = $this->db->prepare('SELECT username FROM users WHERE user_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				$user_id = $card['creator_id'];
				if (!$stmt->bind_param('d', $user_id)) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->store_result();
				$num_of_rows = $stmt->num_rows;
				$stmt->bind_result($username);
			
				// If single username is found store it to card
    			if ($num_of_rows == 1) {
    				$stmt->fetch();
    				$card['creator_name'] = $username;
    			}
					// Get editor name too
   	 		if (isset($card['editor_id']) && !empty($card['editor_id'])) {
   	 			$num_of_rows = 0;
   	 			$user_id = $card['creator_id'];
   	 			$stmt->free_result();
   	 			if (!$stmt->execute()) {
						$this->fail('MySQL execute', $this->db->error);
					}
					$stmt->store_result();
					$num_of_rows = $stmt->num_rows;
					$stmt->bind_result($username);
				
					// If single username is found store it to card
   	 			if ($num_of_rows == 1) {
   	 				$stmt->fetch();
   	 				$card['editor_name'] = $username;
   	 			}
   	 		}
   	 		$stmt->free_result();
   	 		$stmt->close();
    			
    			$card['choices'] = array();
    			if ($include_choices) {
    				$card['choices'] = $this->getCardChoices($card['card_id'], $lang);
    			}
			}
		}
		
//		$mtime = microtime();
//   	$mtime = explode(" ",$mtime);
//   	$mtime = $mtime[1] + $mtime[0];
//   	$endtime = $mtime;
//  	$totaltime = ($endtime - $starttime);
//   	echo "This page was created in ".$totaltime." seconds"; 
   
		return $cards;
	}
	
	/**
	 * Retrieves a single card.
	 * @param double $card_id ID of the card to retrieve.
	 * @param string $lang Language code of the language card is wanted.
    * @return array Containing information about the card.
   */
	public function getCard($card_id, $lang='en-US') {
		
		if (!$stmt = $this->db->prepare('SELECT c.* FROM cards c WHERE c.card_id = ?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($card_id, $card_external_id, $game, $enabled, $repetitive, $created, $modified,
								 $creator_id, $editor_id, $tags, $stat_requirements, $comment, $attachment, $layout,
								 $related_tags, $related_stats, $hidden);
		
		$card = array();
		
		if ($result = $stmt->fetch()) {
			
			$card['card_id'] = $card_id;
			$card['card_external_id'] = $card_external_id;
			if ($card['card_external_id'] < 1) {
				$card['card_external_id'] = $card['card_id'];
			}
			$card['game_id'] = $game;
			$card['enabled'] = $enabled;
			$card['repetitive'] = $repetitive;
			$card['created'] = $created;
			$card['modified'] = $modified;
			$card['creator_id'] = $creator_id;
			$card['editor_id'] = $editor_id;
			
			$tags_array = explode(' ', $tags);
			$clean_tags = array();
				
			foreach ($tags_array as $tag) {
				$trimmed = trim($tag);
				if (!empty($trimmed)) {
					$clean_tags[] = $trimmed;
				}
			}
			$card['tags'] = $clean_tags;
			
			$stats_array = explode(' ', $stat_requirements);
			$clean_stats = array();
				
			foreach ($stats_array as $stat) {
				$trimmed = trim($stat);
				if (!empty($trimmed)) {
					$clean_stats[] = $trimmed;
				}
			}
			$card['stat_requirements'] = $clean_stats;
			$card['comment'] = $comment;
			$card['attachment'] = $attachment;
			$card['layout'] = $layout;
			
			$card['related_tags'] = array();
			$card['related_stats'] = array();
			
			$tags_array = explode(' ', $related_tags);
			$clean_tags = array();
				
			foreach ($tags_array as $tag) {
				$trimmed = trim($tag);
				if (!empty($trimmed)) {
					$clean_tags[] = $trimmed;
				}
			}
			$card['related_tags'] = $clean_tags;
			
			$stats_array = explode(' ', $related_stats);
			$clean_stats = array();
				
			foreach ($stats_array as $stat) {
				$trimmed = trim($stat);
				if (!empty($trimmed)) {
					$clean_stats[] = $trimmed;
				}
			}
			$card['related_stats'] = $clean_stats;
			
			$card['hidden'] = $hidden;
			//$cards[] = $card;
		}
		$stmt->free_result();
		$stmt->close();
		
		if (!empty($card)) {
			
			if (!$stmt = $this->db->prepare('SELECT ct.card_name, ct.description, ct.expert_advice, ct.language_code, ct.is_default FROM card_translations ct WHERE ct.card_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', $card['card_id'])) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($card_name, $description, $expert_advice, $language_code, $is_default);
			
			$card['default_name'] = '';
			$card['default_description'] = '';
			$card['default_expert_advice'] = '';
			$card['default_lang'] = '';
			$card['card_name'] = '';
			$card['description'] = '';
			$card['expert_advice'] = '';
		
			while ($result = $stmt->fetch()) {
				
				if ($language_code == $lang) {
				
					$card['card_name'] = $card_name;
					$card['description'] = $description;
					$card['expert_advice'] = $expert_advice;
					$card['lang'] = $lang;
				}
				if ($is_default) {
					$card['default_name'] = $card_name;
					$card['default_description'] = $description;
					$card['default_expert_advice'] = $expert_advice;
					$card['default_lang'] = $language_code;
				}
			}
			if (!isset($card['lang'])) {
				$card['lang'] = $card['default_lang'];
			}
			$stmt->free_result();
			$stmt->close();
			
			$num_of_rows = 0;

			if (!$stmt = $this->db->prepare('SELECT username FROM users WHERE user_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			$user_id = $card['creator_id'];
			if (!$stmt->bind_param('d', $user_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->store_result();
			$num_of_rows = $stmt->num_rows;
			$stmt->bind_result($username);
			
			// If single username is found store it to card
    		if ($num_of_rows == 1) {
    			$stmt->fetch();
    			$card['creator_name'] = $username;
    		}
			// Get editor name too
    		if (isset($card['editor_id']) && !empty($card['editor_id'])) {
    			$num_of_rows = 0;
    			$user_id = $card['creator_id'];
    			
    			if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->store_result();
				$num_of_rows = $stmt->num_rows;
				$stmt->bind_result($username);
			
				// If single username is found store it to card
    			if ($num_of_rows == 1) {
    				$stmt->fetch();
    				$card['editor_name'] = $username;
    			}
    		}
    		$stmt->free_result();
    		$stmt->close();
    		
    		$card['choices'] = $this->getCardChoices($card['card_id'], $lang);
		}
		return $card;
	}
	
	/**
	 * Enables or disables a card
	 * @param double $card_id ID of the card to update.
	 * @param int $enabled Flag to enable or disable the card. 0: disabled, 1: enabled.
    * @return void
   */
	public function enable_disable_card($card_id, $enabled) {
		
		if (!$stmt = $this->db->prepare('UPDATE cards SET enabled=? WHERE card_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', $enabled, $card_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Updates tags and stats that are related to card.
	 * @param double $card_id ID of the card to update.
    * @return void
   */
	private function updateCardRelatedTagsAndStats($card_id) {
		
		$card = $this->getCard($card_id, 'en-US');
		$card_tags = array();
		$card_stats = array();
		
		if (!empty($card)) {
			
			foreach ($card['tags'] as &$tag) {
				if ($tag[1] != '#') {
					$tag = '#' . $tag;
				}
				$card_tags[] = $tag;
			}
			foreach ($card['stat_requirements'] as $stat) {
				//$card_stats[] = $stat;
				$parsed_stat = GameEngine::parseStat($stat);
				if ($parsed_stat != false) {
					$parsed_stat['key'] = '@' . $parsed_stat['key'];
					$card_stats[] = $parsed_stat['key'];
				}
			}
			unset($stat);
			unset($tag);
			
			// Get stats and tags from results
			if (isset($card['choices'])) {
				foreach ($card['choices'] as $choice) {
					$card_results = $this->getChoiceResults($choice['choice_id'], 'en-US', false);
					foreach ($card_results as $result) {
						$stat_changes = explode("@", $result['stat_changes']);
						if (!empty($stat_changes)) {
							foreach ($stat_changes as $stat) {
								$parsed_stat = GameEngine::parseStat($stat);
								if ($parsed_stat != false) {
									$parsed_stat['key'] = '@' . $parsed_stat['key'];
									// If current stat isn't in array of stats, add it there
									if (!in_array($parsed_stat['key'], $card_stats)) {
										$card_stats[] = $parsed_stat['key'];
									}
								}
							}
						}
						$next_criteria = explode("#", $result['next_criteria']);
						if (!empty($next_criteria)) {
							foreach ($next_criteria as &$crit) {
								$crit = trim($crit);
								if (!empty($crit)) {
									// Check if tag is id
									$pos = strpos($crit, "ID");
									if ($pos !== false) {
										$id = substr($crit, $pos+2);
										if (is_numeric($id)) {
											// and continue if it is
											continue;
										}
									}
									$crit = '#' . $crit;
									// If current tag isn't in array of tags, add it there
									if (!in_array($crit, $card_tags)) {
										$card_tags[] = $crit;
									}
								}
							}
						}
					}
				}
			}
			$related_tags = '';
			$related_stats = '';
			$related_tags = implode(" ", $card_tags);
			$related_tags = trim($related_tags);
			$related_stats = implode(" ", $card_stats);
			$related_stats = trim($related_stats);
			
			if (!$stmt = $this->db->prepare('UPDATE cards SET related_tags=?, related_stats=? WHERE card_id=?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('ssd', $related_tags, $related_stats, $card_id)) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->close();
		}
	}
	
	/**
	 * Handles authentication of the user.
	 * @param string $username Name of the user.
	 * @param string $password Password of the user.
    * @return int ID of the user if successful.
    * Error codes: -1 - user isn't found, -2 - wrong password
   */
	private function authenticateUser($username, $password) {
		
		// Get password hash
		if (!$stmt = $this->db->prepare('SELECT user_id, password FROM users WHERE username=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('s', $username)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->store_result();
		$num_of_rows = $stmt->num_rows;
		$stmt->bind_result($user_id, $password_hash);
		
		$return_value = 0;
		
		// If single password hash isn't found return error -1
    	if ($num_of_rows != 1) {
    		$return_value = -1;
    	}
    	else {
    		$stmt->fetch();
    	}
    	$stmt->free_result();
    	$stmt->close();
		
    	if ($return_value >= 0) {
    		
    		$config = Config::getInstance();
    		if (version_compare($config->php_version, '5.3.7') >= 0) {
    			
    			if (password_verify($password, $password_hash)) {
					// Valid user
    				return $user_id;
				}
				else {
					// Invalid
					$return_value = -2;
				}
			}
			else {
				
				$hasher = new PasswordHash($hash_cost_log2, $hash_portable);
				if ($hasher->CheckPassword($password, $password_hash)) {
					// Valid user
					return $user_id;
				}
				else {
					// Invalid
					$return_value = -2;
				}
				unset($hasher);
			}
    	}
		return $return_value;
	}
	
	/**
	 * Creates a new result.
	 * @param double $choice_id ID of choice of which the new result is related.
	 * @param int $weight The weight of the result. Used to unbalance some results of a same choice.
	 * @param string $stat_changes Changes in player's stats when the result is got.
	 * @param string $next_card_criteria Criteria to use when next card is chosen.
	 * @param int $ordinal Ordinal number of the result.
	 * @param int $choice_ordinal Ordinal number of the choice
    * @return int ID of the new result if successful else -1.
   */
	private function createResult($choice_id, $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal) {
		
		if ($choice_id < 1 && $ordinal < 0) {
			return -1;
		}
		if (!$stmt = $this->db->prepare('INSERT INTO results (choice_id, weight, stat_changes, next_card_criteria, ordinal, choice_ordinal) VALUES (?, ?, ?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dissii', $choice_id, $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Return result_id
		return $this->db->insert_id;
	}
	
	/**
	 * Creates a new choice.
	 * @param double $card_id ID of card of which the new choice is related.
	 * @param int $ordinal Ordinal number of the choice
    * @return int ID of the new choice if successful else -1.
   */
	private function createChoice($card_id, $ordinal) {
		
		if ($card_id < 1 && $ordinal < 0) {
			return -1;
		}
		if (!$stmt = $this->db->prepare('INSERT INTO choices (card_id, ordinal) VALUES (?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('di', $card_id, $ordinal)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Return choice_id
		return $this->db->insert_id;
	}
	
	/**
	 * Updates the ordinal number of a choice.
	 * @param double $choice_id ID of choice of which to update.
	 * @param int $ordinal New ordinal number of the choice.
    * @return void
   */
	private function updateChoiceOrdinal($choice_id, $ordinal) {
		
		if (!$stmt = $this->db->prepare('UPDATE choices SET ordinal=? WHERE choice_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('id', $ordinal, $choice_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Updates the translation of a choice.
	 * @param double $choice_id ID of choice of which translation is updated.
	 * @param string $choice_text Text of the choice in tranlation language.
	 * @param string $lang Language code of the translation.
	 * @param int $is_default Flag to tell if this is the default language of the choice.
	 * @param string $default_lang Language code of default language of the choice.
    * @return void
   */
	private function updateChoiceTranslation($choice_id, $choice_text, $lang, $is_default, $default_lang) {
		
		if (!$stmt = $this->db->prepare('UPDATE choice_translations SET choice_text=?, is_default=? WHERE choice_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sids', $choice_text, $is_default, $choice_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark is_default false to all translations
		if (!$stmt = $this->db->prepare('UPDATE choice_translations set is_default=0 WHERE choice_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $choice_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark right translation to default
		if (!$stmt = $this->db->prepare('UPDATE choice_translations set is_default=1 WHERE choice_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $choice_id, $default_lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Updates result.
	 * @param double $result_id ID of result to update.
	 * @param int $weight The weight of the result. Used to unbalance some results of a same choice.
	 * @param string $stat_changes Changes in player's stats when the result is got.
	 * @param string $next_card_criteria Criteria to use when next card is chosen.
	 * @param int $ordinal Ordinal number of the result.
	 * @param int $choice_ordinal Ordinal number of the choice
    * @return void
   */
	private function updateResult($result_id, $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal) {
		
		if (!$stmt = $this->db->prepare('UPDATE results SET weight=?, stat_changes=?, next_card_criteria=?, ordinal=?, choice_ordinal=? WHERE result_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('issiid', $weight, $stat_changes, $next_card_criteria, $ordinal, $choice_ordinal, $result_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Updates the translation of a result.
	 * @param double $result_id ID of result of which translation is updated.
	 * @param string $result_text Text of the result in tranlation language.
	 * @param string $lang Language code of the translation.
	 * @param int $is_default Flag to tell if this is the default language of the result.
	 * @param string $default_lang Language code of default language of the result.
    * @return void
   */
	private function updateResultTranslation($result_id, $result_text, $lang, $is_default, $default_lang) {
		
		if (!$stmt = $this->db->prepare('UPDATE result_translations SET result_text=?, is_default=? WHERE result_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('sids', $result_text, $is_default, $result_id, $lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark is_default false to all translations
		if (!$stmt = $this->db->prepare('UPDATE result_translations set is_default=0 WHERE result_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', $result_id)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		// Mark right translation to default
		if (!$stmt = $this->db->prepare('UPDATE result_translations set is_default=1 WHERE result_id=? AND language_code=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('ds', $result_id, $default_lang)) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
	}
	
	/**
	 * Retrieves the choices of a card.
	 * @param double $card_id ID of card of which choices are retrieved.
	 * @param string $lang Language code of the choices.
    * @return array Contains choices.
   */
	private function getCardChoices($card_id, $lang='en-US') {
		
		if (!$stmt = $this->db->prepare('SELECT c.* FROM choices c WHERE c.card_id = ? ORDER BY c.ordinal')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		
		if (!$stmt->bind_param('d', doubleval($card_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($choice_id, $card, $ordinal);
		
		$choices = array();
		
		while ($result = $stmt->fetch()) {
			
			$choice = array();
			$choice['choice_id'] = $choice_id;
			$choice['card_id'] = $card;
			$choice['ordinal'] = $ordinal;
			$choices[] = $choice;
		}
		$stmt->free_result();
		$stmt->close();
		
		foreach ($choices as &$choice) {
		
			if (!$stmt = $this->db->prepare('SELECT ct.choice_text, ct.language_code, ct.is_default FROM choice_translations ct WHERE ct.choice_id = ?')) {
				$this->fail('MySQL prepare', $this->db->error);
			}
			if (!$stmt->bind_param('d', doubleval($choice['choice_id']))) {
				$this->fail('MySQL bind_param', $this->db->error);
			}
			if (!$stmt->execute()) {
				$this->fail('MySQL execute', $this->db->error);
			}
			$stmt->bind_result($choice_text, $language_code, $is_default);
			
			$choice['choice_text'] = '';
			$choice['lang'] = $lang;
		
			while ($result = $stmt->fetch()) {
				
				if ($language_code == $lang) {
				
					$choice['choice_text'] = $choice_text;
				}
				if ($is_default) {
					$choice['default_text'] = $choice_text;
					$choice['default_lang'] = $language_code;
				}
			}
			$stmt->free_result();
			$stmt->close();
		}
		
		return $choices;
	}
	
	/**
	 * Creates a new sudden scenario.
	 * @param double $game_id ID of the game of which scenario is added.
	 * @param int $odds The odds for the scenario to occur.
	 * @param string $result Criteria for card to show.
	 * @param int $ordinal Ordinal number of the sudden scenario.
    * @return int ID of the new sudden scenario.
   */
	private function createSuddenScenario($game_id, $odds, $result, $ordinal) {
		
		if (!$stmt = $this->db->prepare('INSERT INTO sudden_scenarios (game_id, odds, result, ordinal) VALUES (?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('disi', doubleval($game_id), intval($odds), $result, intval($ordinal))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Creates a new ending criteria.
	 * @param double $game_id ID of the game of which criteria is added.
	 * @param string $result Criteria for card to show.
	 * @param int $ordinal Ordinal number of the ending criteria.
    * @return int ID of the new ending criteria.
   */
	private function createEndingCriteria($game_id, $result, $ordinal) {
		
		if (!$stmt = $this->db->prepare('INSERT INTO ending_criteria (game_id, result, ordinal) VALUES (?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dsi', doubleval($game_id), $result, intval($ordinal))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Creates a new condition for sudden scenario or ending criteria.
	 * @param double $object_id ID of the ending criteria or sudden scenario
	 * of which new condition is added.
	 * @param string $condition_text Stat requirements for the condition.
	 * @param string $condition_logic Logic of the condition so that creates logical chain.
	 * Possible values 'AND', 'OR' and empty ''. 
	 * @param int $ordinal Ordinal number of the condition.
	 * @param $table_name Tells whether condition is for sudden scenarios or ending criteria.
	 * Possible values are 'ending_conditions' and 'scenario_conditions'.
    * @return int ID of the new condition
   */
	private function createCondition($object_id, $condition_text, $condition_logic, $ordinal, $table_name) {
		
		$object_name = '';
		// Choose the right name for object
		if ($table_name == 'ending_conditions') {
			$object_name = 'criteria';
		}
		else {
			$object_name = 'scenario';
		}
		$object_name .= '_id';
		
		if (!$stmt = $this->db->prepare('INSERT INTO ' . $table_name . ' (' . $object_name . ', condition_text, condition_logic, ordinal) VALUES (?, ?, ?, ?)')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('dssi', doubleval($object_id), $condition_text, $condition_logic, intval($ordinal))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->close();
		
		return $this->db->insert_id;
	}
	
	/**
	 * Deletes all sudden scenarios of a game.
	 * @param double $game_id ID of the game of which scenarios to delete.
    * @return void
   */
	private function deleteSuddenScenarios($game_id) {
		
		// Get scenario_ids
		if (!$stmt = $this->db->prepare('SELECT scenario_id FROM sudden_scenarios WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($scenario_id);
	
		$scenarios = array();
	
		while ($result = $stmt->fetch()) {
			$scenarios[] = $scenario_id;
		}
		$stmt->close();
		// Delete scenarios
		if (!$stmt = $this->db->prepare('DELETE FROM sudden_scenarios WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		
		$stmt->close();
		
		if (count($scenarios) > 0) {
			
			foreach ($scenarios as $scenario) {
				// Delete scenario conditions
				if (!$stmt = $this->db->prepare('DELETE FROM scenario_conditions WHERE scenario_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($scenario))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->close();
			}
		}
	}
	
	/**
	 * Deletes all ending criteria of a game.
	 * @param double $game_id ID of the game of which criteria to delete.
    * @return void
   */
	private function deleteEndingCriteria($game_id) {
		
		// Get criteria_ids
		if (!$stmt = $this->db->prepare('SELECT criteria_id FROM ending_criteria WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($criteria_id);
	
		$criterias = array();
	
		while ($result = $stmt->fetch()) {
			$criterias[] = $criteria_id;
		}
		$stmt->close();
		// Delete criteria
		if (!$stmt = $this->db->prepare('DELETE FROM ending_criteria WHERE game_id=?')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$result = $this->db->affected_rows;
		
		$stmt->close();
		
		if (count($criterias) > 0) {
			
			foreach ($criterias as $criteria) {
				// Delete criteria conditions
				if (!$stmt = $this->db->prepare('DELETE FROM ending_conditions WHERE criteria_id=?')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($criteria))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->close();
			}
		}
	}
	
	/**
	 * Retrieves all sudden scenarios of a game.
	 * @param double $game_id ID of the game of which scenarios to retrieve.
    * @return array An array containing sudden scenarios.
   */
	private function getSuddenScenarios($game_id) {
		
		// Get scenarios
		if (!$stmt = $this->db->prepare('SELECT scenario_id, odds, result, ordinal FROM sudden_scenarios WHERE game_id=? ORDER BY ordinal ASC')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($scenario_id, $odds, $result, $ordinal);
	
		$scenarios = array();
	
		while ($res = $stmt->fetch()) {
			$scenario = array();
			$scenario['scenario_id'] = $scenario_id;
			$scenario['odds'] = $odds;
			$scenario['result'] = $result;
			$scenario['ordinal'] = $ordinal;
			$scenarios[] = $scenario;
		}
		$stmt->close();
		
		if (count($scenarios) > 0) {
			
			foreach ($scenarios as &$scenario) {
				// Get conditions
				if (!$stmt = $this->db->prepare('SELECT condition_id, condition_text, condition_logic, ordinal FROM scenario_conditions WHERE scenario_id=? ORDER BY ordinal ASC')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($scenario['scenario_id']))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($condition_id, $condition_text, $condition_logic, $ordinal);
	
				$conditions = array();
	
				while ($res = $stmt->fetch()) {
					$condition = array();
					$condition['condition_id'] = $condition_id;
					$condition['condition_text'] = $condition_text;
					$condition['condition_logic'] = $condition_logic;
					$condition['ordinal'] = $ordinal;
					$conditions[] = $condition;
				}
				$stmt->close();
				$scenario['conditions'] = $conditions;
			}
		}
		return $scenarios;
	}
	
	/**
	 * Retrieves all ending criteria of a game.
	 * @param double $game_id ID of the game of which criteria to retrieve.
    * @return array An array containing ending criteria.
   */
	private function getEndingCriteria($game_id) {
		
		// Get ending criteria
		if (!$stmt = $this->db->prepare('SELECT criteria_id, result, ordinal FROM ending_criteria WHERE game_id=? ORDER BY ordinal ASC')) {
			$this->fail('MySQL prepare', $this->db->error);
		}
		if (!$stmt->bind_param('d', doubleval($game_id))) {
			$this->fail('MySQL bind_param', $this->db->error);
		}
		if (!$stmt->execute()) {
			$this->fail('MySQL execute', $this->db->error);
		}
		$stmt->bind_result($criteria_id, $result, $ordinal);
	
		$criterias = array();
	
		while ($res = $stmt->fetch()) {
			$criteria = array();
			$criteria['criteria_id'] = $criteria_id;
			$criteria['result'] = $result;
			$criteria['ordinal'] = $ordinal;
			$criterias[] = $criteria;
		}
		$stmt->close();
		
		if (count($criterias) > 0) {
			
			foreach ($criterias as &$criteria) {
				
				if (!$stmt = $this->db->prepare('SELECT condition_id, condition_text, condition_logic, ordinal FROM ending_conditions WHERE criteria_id=? ORDER BY ordinal ASC')) {
					$this->fail('MySQL prepare', $this->db->error);
				}
				if (!$stmt->bind_param('d', doubleval($criteria['criteria_id']))) {
					$this->fail('MySQL bind_param', $this->db->error);
				}
				if (!$stmt->execute()) {
					$this->fail('MySQL execute', $this->db->error);
				}
				$stmt->bind_result($condition_id, $condition_text, $condition_logic, $ordinal);
	
				$conditions = array();
	
				while ($res = $stmt->fetch()) {
					$condition = array();
					$condition['condition_id'] = $condition_id;
					$condition['condition_text'] = $condition_text;
					$condition['condition_logic'] = $condition_logic;
					$condition['ordinal'] = $ordinal;
					$conditions[] = $condition;
				}
				$stmt->close();
				$criteria['conditions'] = $conditions;
			}
		}
		return $criterias;
	}
}

?>