<?php
/**
*	@file functions.php
*	@brief Contains common functions of GAL2.
*/
require_once(dirname(__FILE__) . '/settings.php');

/**
* Function to autoload all classes (legacy).
*/
function my_autoloader($class) {
	include 'classes/' . $class . '.class.php';
}
spl_autoload_register('my_autoloader');

/**
* Define the url path for the localization resources
*/
defined('LOCALE_PATH') or define('LOCALE_PATH', dirname(__FILE__) . '/../locales/');

/**
* Define the language using language code based on BCP 47 + RFC 4644,
* http://www.rfc-editor.org/rfc/bcp/bcp47.txt
*
* The language files can be found in directory 'locales'
*/

/**
* Load the proper language file and return the translated phrase
*
* The language file is JSON encoded and returns an associative array
* Language filename is determined by BCP 47 + RFC 4646
* http://www.rfc-editor.org/rfc/bcp/bcp47.txt
*
* @param string $phrase The phrase that needs to be translated
* @return string
*/
function localize($phrase) {
	
	global $language;
	global $default_language;
	
   /* Static keyword is used to ensure the file is loaded only once */
   static $translations = NULL;
   static $default_translations = NULL;
   /* If no instance of $translations has occured load the language file */
   if (is_null($translations)) {
		$lang_file = LOCALE_PATH . $language . '.txt';
		/*if (!file_exists($lang_file)) {
			$lang_file = LOCALE_PATH . 'en-US.txt';
		}*/
		$lang_file_content = file_get_contents($lang_file, FILE_USE_INCLUDE_PATH);
		/* Load the language file as a JSON object and transform it into an associative array */
		$translations = json_decode($lang_file_content, true);
	}
	// Same thing to default language
   if (is_null($default_translations)) {
		$lang_file = LOCALE_PATH . $default_language . '.txt';
		$lang_file_content = file_get_contents($lang_file, FILE_USE_INCLUDE_PATH);
		$default_translations = json_decode($lang_file_content, true);
	}
	// Check whether translation is found in given language. if not use default language
	if (!isset($translations[$phrase])) {
		return $default_translations[$phrase];
	}
	return $translations[$phrase];
}

/**
* Function to do all common operations in start of every page.
* Handles starting session, setting timezone, setting internal encoding and assigning current language.
* @return void
*/
function do_basic_operations() {
	
	session_start();
	date_default_timezone_set("Europe/Helsinki");
	mb_internal_encoding("UTF-8");
	
	// Parameter to tell whether session has been expired
	if (!isset($_SESSION['active'])) {
		$_SESSION['active'] = true;
	}
	
	global $language;
	
	if (isset($_REQUEST['lang']) && !empty($_REQUEST['lang'])) {
		$_SESSION["lang"] = $_REQUEST['lang'];
	}
	if (isset($_SESSION["lang"]) && !empty($_SESSION["lang"])) {
		$language = $_SESSION["lang"];
	}
	if (isset($_GET['game_slug']) && !empty($_GET['game_slug'])) {
		$_SESSION["slug"] = $_GET['game_slug'];
	}
	elseif (isset($_SESSION["slug"])) {
		$_GET['game_slug'] = $_SESSION["slug"];
	}
	
	$_SESSION['currentview'] = $_SERVER['PHP_SELF'];
}

/**
* Function to tell if mod_rewrite is enabled.
* @return bool Is it enabled?
*/
function modRewriteEnabled() {
	
	$mod_rewrite = FALSE;
	if (function_exists("apache_get_modules")) {
   	$modules = apache_get_modules();
   	$mod_rewrite = in_array("mod_rewrite",$modules);
	}
	if (!isset($mod_rewrite) && isset($_SERVER["HTTP_MOD_REWRITE"])) {
   	$mod_rewrite = ($_SERVER["HTTP_MOD_REWRITE"]=="on" ? TRUE : FALSE); 
	}
	else {
  		$mod_rewrite = getenv('HTTP_MOD_REWRITE')=='on' ? true : false ;
	}
	if (!isset($mod_rewrite)) {
		$config = Config::getInstance();
		$slug = '';
		if (isset($_GET['game_slug']) && !empty($_GET['game_slug'])) {
			$slug = $_GET['game_slug'];
		}
   	// last solution; call a specific page as "mod-rewrite" have been enabled; based on result, we decide.
   	$result = file_get_contents($config->baseurl . "game/" . $slug . "/test_mod_rewrite.php");
   	$mod_rewrite  = ($result=="ok" ? TRUE : FALSE);
	}
	return $mod_rewrite;
}

/**
* Function to format application links based on if mod_rewrite is enabled or not.
* @param string $url Url to format. Only urls that point to "app" folder are formatted.
* @return string Formatted url.
*/
function formatAppLink($url) {
	
	if (modRewriteEnabled()) {
		
		$config = Config::getInstance();
		$new_url = $config->baseurl;
		// First check if app folder is found
		$app = strpos($url, "app/");
		if ($app === false) {
			return $url;
		}
		// Remove the beginning of url
		$temp = str_replace(($config->baseurl . 'app/'), '', $url);
		$file = '';
		// get file
		$file_end = strpos($temp, ".php");
		if ($file_end !== false) {
			$file = substr($temp, 0, $file_end+4);
		}
		//	Remove file from temp
		$temp = str_replace($file, '', $temp);
		// Remove file extension
		$file = str_replace('.php', '', $file);
		// get query
		$query_start = strpos($temp, "?");
		$query_string = '';
		if ($query_start !== false) {
			$query_string = substr($temp, $query_start);
		}
		// Remove query from temp
		$temp = str_replace($query_string, '', $temp);
		
		$query_string = str_replace('?', '', $query_string);
		$query_params = explode('&', $query_string);
		
		$slug = '';
		$new_query = '';
		// get current game and clean query params
		foreach ($query_params as &$param) {
			$param = str_replace('&', '', $param);
			$pos = strpos($param, "game_slug=");
			if ($pos !== false) {
				$slug = str_replace('game_slug=', '', $param);
			}
			else {
				$new_query .= '&' . $param;
			}
		}
		if (strlen($new_query) > 0) {
			$new_query[0] = '?';
		}
		if ($file === 'index') {
			$file = '';
		}
		else {
			$file = '/' . $file;
		}
		$new_url .= 'game/' . $slug . $file . $new_query;
		return $new_url;
	}
	return $url;
}

/**
* Function to print the end of every page.
* @return void
*/
function print_html_end() {
?>
</body>
</html>
<?php 
}

/**
* Function to print the language bar.
* @param string $active_lang Tells which language is currently selected (optionally).
* @return void
*/
function print_language_bar($active_lang='') {

	$config = Config::getInstance();
	$slug = '';
	if (isset($_SESSION['slug'])) {
		$slug = $_SESSION['slug'];
	}
	$current_file_path = $_SESSION['currentview'];
	$parts = explode('/', $current_file_path);
	$current = $parts[count($parts) - 1];
	
	global $languages;
	$counter = 0;
	foreach ($languages as $lang) {
		if ($counter > 0) {
			echo(' | ');
		}
		$counter++;
		// Get the correct language abbreviation
		$lang_file = LOCALE_PATH . $lang . '.txt';
		$lang_file_content = file_get_contents($lang_file, FILE_USE_INCLUDE_PATH);
		/* Load the language file as a JSON object and transform it into an associative array */
		$texts = json_decode($lang_file_content, true);
		$abbreviation = $texts['LANG_ABBREVIATION'];
		
		if ($active_lang != $lang) { ?>
		<a href="<?php echo formatAppLink($config->baseurl . 'app/' . $current . '?game_slug=' . $slug . '&lang=' . $lang); ?>"><?php echo($abbreviation);?></a>
<?php } else { echo($abbreviation); }
	}
}

/**
* Function to print the game name into header.
* @return void
*/
function print_game_name() {

	$game_name = '';

	if (isset($_SESSION['game']['name'])) {
		$game_name = $_SESSION['game']['name'];
	}
	if (!empty($game_name)) { ?>
				<?php //echo(localize('GAME')); ?>
				<?php //echo $game_name; ?>
				<dt><?php echo(localize('GAME')); ?></dt>
				<dd><?php echo $game_name; ?></dd>
<?php	} 
}

/**
* Function to print login information and links.
* @return void
*/
function print_login($language_bar, $show_user=1, $show_registration=1) {
	global $language;
	$current_lang = $language;
	if (isset($_SESSION['lang'])) {
		$current_lang = $_SESSION['lang'];
	}
	
	$username = '';
	if (isset($_SESSION['user']['username'])) {
		$username = $_SESSION['user']['username'];
	}
	$isLogged = false;
	if (isset($_SESSION['isLogged']) && $_SESSION['isLogged']) {
		$isLogged = true;
	}
	$config = Config::getInstance();
	$slug = '';
	if (isset($_SESSION['slug'])) {
		$slug = $_SESSION['slug'];
	}
	$show_language_bar = 1;
	if (isset($_SESSION['game']['multiple_languages'])) {
		$show_language_bar = $_SESSION['game']['multiple_languages'];
	}
	else if (isset($language_bar)) {
		$show_language_bar = $language_bar;
	}
	$show_user_name = 1;
	if (isset($_SESSION['game']['use_name'])) {
		$show_user_name = $_SESSION['game']['use_name'];
	}
	else if (isset($show_user)) {
		$show_user_name = $show_user;
	}
	$show_registration_link = 1;
	if (isset($_SESSION['game']['use_login'])) {
		$show_registration_link = $_SESSION['game']['use_login'];
	}
	else if (isset($show_registration)) {
		$show_registration_link = $show_registration;
	}
	?>
		<?php /*if (!empty($username) && $show_user_name) :*/?>
			<?php /*echo(localize("LOGIN_USER")); */?>
			<?php /*if ($isLogged) : echo('<a class="jqModal" id="user_settings_button" href="#">'. $username . '</a>'); else : echo $username; endif; */?>
		<?php /*endif;*/ ?>
		<?php /*if ($show_language_bar) : ?><?php echo localize("LANGUAGE"); ?> <?php print_language_bar($current_lang); ?><br /><?php endif;*/ ?>
		<?php /*if ($isLogged) :?>
			<!--<a class="log_out" href="<?php //echo formatAppLink($config->baseurl . 'general/logout.php?game_slug=' . $slug); ?>"><?php //echo localize("LOGOUT"); ?></a><br />-->
		<?php /*else: ?>
		<?php 	if (!empty($username)) :*/?>
			<!--<a class="log_out" href="<?php echo formatAppLink($config->baseurl . 'general/logout.php?game_slug=' . $slug); ?>"><?php echo localize("QUIT"); ?></a><br />-->
		<?php 		/*if ($show_registration_link) : */?>
			<!--<a class="register" href="<?php echo formatAppLink($config->baseurl . 'app/index.php?game_slug=' . $slug . '&showpanel=app_index_register'); ?>"><?php echo localize("REGISTRATION"); ?></a><br />-->
		<?php 	/*	endif;*/?>

<?php if (!empty($username) && $show_user_name) :?>
			<dt><?php echo(localize("LOGIN_USER")); ?></dt>
			<dd><?php if ($isLogged) : echo('<a class="jqModal" id="user_settings_button" href="#">'. $username . '</a>'); else : echo $username; endif; ?></dd>
		<?php endif; ?>
		<?php if ($show_language_bar) : ?><dt><?php echo localize("LANGUAGE"); ?></dt><dd class="selected_language"><?php print_language_bar($current_lang); ?></dd><?php endif; ?>
		<?php if ($isLogged) :?>
			<dt><a class="log_out" href="<?php echo formatAppLink($config->baseurl . 'general/logout.php?game_slug=' . $slug); ?>"><?php echo localize("LOGOUT"); ?></a></dt><dd></dd>
		<?php else: ?>
		<?php 	if (!empty($username)) :?>
			<dt><a class="log_out" href="<?php echo formatAppLink($config->baseurl . 'general/logout.php?game_slug=' . $slug); ?>"><?php echo localize("QUIT"); ?></a></dt><dd></dd>
		<?php 		if ($show_registration_link) : ?>
			<dt><a class="register" href="<?php echo formatAppLink($config->baseurl . 'app/index.php?game_slug=' . $slug . '&showpanel=app_index_register'); ?>"><?php echo localize("REGISTRATION"); ?></a></dt><dd></dd>
		<?php 		endif;?>
		<?php 	endif;?>
		<?php endif; ?>
		<?php 
				$current_file_path = $_SESSION['currentview'];
				$parts = explode('/', $current_file_path);
				$current = $parts[count($parts) - 1];
				$pos = strpos($current, '#');
				if ($pos !== false) {
					$current = substr($current, 0, $pos);
				}
				if ($current == 'simulation.php' || $current == 'simulation' || $current == 'end' || $current == 'end.php') :
		?>
       <!--<a href="<?php echo formatAppLink($config->baseurl . 'app/index.php?game_slug=' . $slug); ?>"><?php echo localize("BACK_TO_FRONTPAGE"); ?></a><br />-->
    	<dt style="width:100%"><a href="<?php echo formatAppLink($config->baseurl . 'app/index.php?game_slug=' . $slug); ?>"><?php echo localize("BACK_TO_FRONTPAGE"); ?></a></dt>  
	<?php endif; ?>
<?php
}

/**
* Function to generate unique filename.
* @param string $filename Filename that gets modified if there is one with the same name already.
* @param string $path Filepath to where file is supposed to be created.
* @return string Unique filename string.
*/
function generateFilename($filename, $path) {
	
	$new_filename = $filename;
	if (isFile($path,$filename)) {
		
		$ext = findExtension($filename);
		$extIndx = strrpos($filename,'.');
		$basename = substr($filename,0,$extIndx);
		$nbr = 2;
		$new_filename = $basename . strval($nbr) . $ext;
		
		while (isFile($path,$new_filename)) {
			
			$nbr++;
			$new_filename = $basename . strval($nbr) . $ext;
		}
	}
	
	return $new_filename;
}

/**
* Function to check whether or not file exists.
* @param string $path Filepath to where to check.
* @param string $filename Filename of file to check.
* @return bool Is file found or not?
*/
function isFile($path, $filename) {
	
	//using the opendir function
	$dir_handle = @opendir($path);
	if (!$dir_handle) {
		return false;
	}
	
	while ($file = readdir($dir_handle)) {
    	if ($file!="." && $file!="..") {
			if ($file == $filename) {
				//closing the directory
				closedir($dir_handle);
      		return true;
			}
    	}
  	}
	//closing the directory
	closedir($dir_handle);
	return false;
}

/**
* Function to find file extension.
* @param string $str Filename which to check.
* @return string File extension.
*/
function findExtension($str) {
	
	$extIndx = strrpos($str,'.');
	$ext = substr($str,$extIndx);
	
	return $ext;
}

/**
* Function to clean the given url to usable.
* Taken from http://neo22s.com/slug/
* @param string $url Url to format.
* @return string Cleaned url.
*/
function friendly_url($url) {
	
	//replace accent characters, depends your language is needed
	$url = replace_accents($url);
	// everything to lower and no spaces begin or end
	$url = strtolower(trim($url));

	// decode html maybe needed if there's html I normally don't use this
	//$url = html_entity_decode($url,ENT_QUOTES,'UTF8');
	
	// adding - for spaces and union characters
	$find = array(' ', '&', '\r\n', '\n', '+',',');
	$url = str_replace ($find, '-', $url);
	
	//delete and replace rest of special chars
	$find = array('/[^a-z0-9\-<>_]/', '/[\-]+/', '/<[^>]*>/');
	$repl = array('', '-', '');
	$url = preg_replace ($find, $repl, $url);
	
	//return the friendly url
	return $url;
}

/**
* Function to replace accents characters.
* Taken from http://neo22s.com/slug/
* @param string $var String to format.
* @return string Where accents are removed.
*/
function replace_accents($var){
	//replace for accents catalan spanish and more
	$a = array('À', 'Á', 'Â', 'Ã', 'Ä', 'Å', 'Æ', 'Ç', 'È', 'É', 'Ê', 'Ë', 'Ì', 'Í', 'Î', 'Ï', 'Ð', 'Ñ', 'Ò', 'Ó', 'Ô', 'Õ', 'Ö', 'Ø', 'Ù', 'Ú', 'Û', 'Ü', 'Ý', 'ß', 'à', 'á', 'â', 'ã', 'ä', 'å', 'æ', 'ç', 'è', 'é', 'ê', 'ë', 'ì', 'í', 'î', 'ï', 'ñ', 'ò', 'ó', 'ô', 'õ', 'ö', 'ø', 'ù', 'ú', 'û', 'ü', 'ý', 'ÿ', 'Ā', 'ā', 'Ă', 'ă', 'Ą', 'ą', 'Ć', 'ć', 'Ĉ', 'ĉ', 'Ċ', 'ċ', 'Č', 'č', 'Ď', 'ď', 'Đ', 'đ', 'Ē', 'ē', 'Ĕ', 'ĕ', 'Ė', 'ė', 'Ę', 'ę', 'Ě', 'ě', 'Ĝ', 'ĝ', 'Ğ', 'ğ', 'Ġ', 'ġ', 'Ģ', 'ģ', 'Ĥ', 'ĥ', 'Ħ', 'ħ', 'Ĩ', 'ĩ', 'Ī', 'ī', 'Ĭ', 'ĭ', 'Į', 'į', 'İ', 'ı', 'Ĳ', 'ĳ', 'Ĵ', 'ĵ', 'Ķ', 'ķ', 'Ĺ', 'ĺ', 'Ļ', 'ļ', 'Ľ', 'ľ', 'Ŀ', 'ŀ', 'Ł', 'ł', 'Ń', 'ń', 'Ņ', 'ņ', 'Ň', 'ň', 'ŉ', 'Ō', 'ō', 'Ŏ', 'ŏ', 'Ő', 'ő', 'Œ', 'œ', 'Ŕ', 'ŕ', 'Ŗ', 'ŗ', 'Ř', 'ř', 'Ś', 'ś', 'Ŝ', 'ŝ', 'Ş', 'ş', 'Š', 'š', 'Ţ', 'ţ', 'Ť', 'ť', 'Ŧ', 'ŧ', 'Ũ', 'ũ', 'Ū', 'ū', 'Ŭ', 'ŭ', 'Ů', 'ů', 'Ű', 'ű', 'Ų', 'ų', 'Ŵ', 'ŵ', 'Ŷ', 'ŷ', 'Ÿ', 'Ź', 'ź', 'Ż', 'ż', 'Ž', 'ž', 'ſ', 'ƒ', 'Ơ', 'ơ', 'Ư', 'ư', 'Ǎ', 'ǎ', 'Ǐ', 'ǐ', 'Ǒ', 'ǒ', 'Ǔ', 'ǔ', 'Ǖ', 'ǖ', 'Ǘ', 'ǘ', 'Ǚ', 'ǚ', 'Ǜ', 'ǜ', 'Ǻ', 'ǻ', 'Ǽ', 'ǽ', 'Ǿ', 'ǿ'); 
	$b = array('A', 'A', 'A', 'A', 'A', 'A', 'AE', 'C', 'E', 'E', 'E', 'E', 'I', 'I', 'I', 'I', 'D', 'N', 'O', 'O', 'O', 'O', 'O', 'O', 'U', 'U', 'U', 'U', 'Y', 's', 'a', 'a', 'a', 'a', 'a', 'a', 'ae', 'c', 'e', 'e', 'e', 'e', 'i', 'i', 'i', 'i', 'n', 'o', 'o', 'o', 'o', 'o', 'o', 'u', 'u', 'u', 'u', 'y', 'y', 'A', 'a', 'A', 'a', 'A', 'a', 'C', 'c', 'C', 'c', 'C', 'c', 'C', 'c', 'D', 'd', 'D', 'd', 'E', 'e', 'E', 'e', 'E', 'e', 'E', 'e', 'E', 'e', 'G', 'g', 'G', 'g', 'G', 'g', 'G', 'g', 'H', 'h', 'H', 'h', 'I', 'i', 'I', 'i', 'I', 'i', 'I', 'i', 'I', 'i', 'IJ', 'ij', 'J', 'j', 'K', 'k', 'L', 'l', 'L', 'l', 'L', 'l', 'L', 'l', 'l', 'l', 'N', 'n', 'N', 'n', 'N', 'n', 'n', 'O', 'o', 'O', 'o', 'O', 'o', 'OE', 'oe', 'R', 'r', 'R', 'r', 'R', 'r', 'S', 's', 'S', 's', 'S', 's', 'S', 's', 'T', 't', 'T', 't', 'T', 't', 'U', 'u', 'U', 'u', 'U', 'u', 'U', 'u', 'U', 'u', 'U', 'u', 'W', 'w', 'Y', 'y', 'Y', 'Z', 'z', 'Z', 'z', 'Z', 'z', 's', 'f', 'O', 'o', 'U', 'u', 'A', 'a', 'I', 'i', 'O', 'o', 'U', 'u', 'U', 'u', 'U', 'u', 'U', 'u', 'U', 'u', 'A', 'a', 'AE', 'ae', 'O', 'o'); 
	$var = str_replace($a, $b,$var);
	return $var; 
}

/**
* Function to replace dynamic texts and values in string.
* @param string $text_to_process Text to be processed.
* @param array $stats Stats from which conditions are compared.
* @return string Modified text.
*/
function generate_dynamic_text($text_to_process, $stats) {
	
	$modified_text = $text_to_process;
	
	$stat_names = array_keys($stats);
	$stat_values = array_values($stats);
	
	// Parse direct stat values
	for ($i = 0; $i < count($stat_names); $i++) {
		$modified_text = str_replace(PLAYER_STAT_TAG_START . $stat_names[$i] . PLAYER_STAT_TAG_END, str_replace('=', '', $stat_values[$i]), $modified_text);
	}		
	// Parse conditional texts
	$searching = true;
	$test = '';
	while ($searching) {
		$start_index = -1;
		$end_index = -1;
		// Find first occurrence of starting tag
		$start_index = strpos($modified_text, PLAYER_CONDITIONAL_TEXT_TAG_START);
		
		if ($start_index === false) {
			$searching = false;
			continue;
		}
		// Find first occurrence of ending tag
		$end_index = strpos($modified_text, PLAYER_CONDITIONAL_TEXT_TAG_END, $start_index+1);
		if ($end_index === false) {
			$searching = false;
			continue;
		}
		// Get conditional string from the text
		$temp = trim(substr($modified_text, $start_index + strlen(PLAYER_CONDITIONAL_TEXT_TAG_START), $end_index - $start_index - strlen(PLAYER_CONDITIONAL_TEXT_TAG_END)));
		// Remove conditional string from the text
		$modified_text = substr_replace($modified_text, '', $start_index, $end_index - $start_index + strlen(PLAYER_CONDITIONAL_TEXT_TAG_END));
		// Get condition and text
		if ($temp[0] != '[' || $temp[strlen($temp)-1] != ']') {
			continue;
		}
		$temp = substr($temp, 1, strlen($temp)-2);
		$parts = explode('][', $temp);
		if (count($parts) != 2) {
			continue;
		}
		$condition = $parts[0];
		// Replace lesser than and greater than with equivalent symbols
		$condition = str_replace('&gt;', '>', $condition);
		$condition = str_replace('&lt;', '>', $condition);
		
		$conditional_text = $parts[1];

		$is_valid = true;
		$conditions = explode(' ', $condition);
		
		foreach ($conditions as $stat) {
						
			$parsed_stat = GameEngine::parseStat($stat);

			if ($parsed_stat !== false) {
				
				$value_to_compare;
				// If stat exists
				if (isset($stats[$parsed_stat['key']])) {
					$value_to_compare = $stats[$parsed_stat['key']];
				}
				else {
					$value_to_compare = 0;
				}
				$is_valid = GameEngine::validateStat($parsed_stat, $value_to_compare);
				if (!$is_valid) {
					break;
				}
			}
			else {
				$is_valid = false;
			}
		}
		// If condition is valid add conditional text to string
		if ($is_valid) {
			$modified_text = substr_replace($modified_text, $conditional_text, $start_index, 0);
		}
	}
	return $modified_text;
}

/**
 * Function to replace player name tags in given text.
 * @param string $text Text to modify.
 * @param double $profile_id ID of player profile.
 */
function replace_player_name($text, $profile_id) {
	
	$text_to_modify = $text;
	$db = DB::getInstance();
	$profile = $db->getProfile($profile_id);
	
	if (!empty($profile)) {
		$text_to_modify = str_replace(PLAYER_NAME_TAG, $profile['player_name'], $text);
	}
	return $text_to_modify;
}